#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 09:31:52 2024

@author: lorenzo piu
"""

"""
# Cut a 3D scalar

The goal of this exercise is to explain how the class Scalar works, in the 
two modes available. Once understood this, using this class should simplify
operations with large 3d arrays.

First of all, we need to import the module. Make sure you installed
the package with the command
pip install aPrioriDNS
"""
import aPriori as ap
import numpy as np

"""
now let's define a generic array that we will use as a scalar field """

shape = [30, 20, 15]
array = np.random.rand(*shape)

scalar = ap.Scalar(shape=shape, value=array)
print(f'Initial shape:\n{scalar.shape}')
"""
Cut the field using the mode 'equal'. This is useful when we want to
cut the extrema after filtering operations, or in general after operations 
that have problems in the boundary treatment.
Let's say we want to cut two cells from every boundary:
"""
scalar.cut(n_cut=2, mode='equal')
print(f'Scalar\'s shape after cutting with \'equal\' mode:\n{scalar.shape}')

"""
This time we are going to download the DNS dataset, to use the Scalar class
in a more useful way. This approach uses a pointer to the file instead than
loading the data. This can help when processing large datasets, avoiding to 
overload the RAM and avoid kernel restarting.
If you downloaded a different folder from Github or from Blastnet, 
it will be enough to change the directory path and comment the line  
ap.download()
"""

import os

ap.download() # Downloads the dataset at https://github.com/LorenzoPiu/aPrioriDNS/tree/main/data
# The download can take a couple of minutes

directory = os.path.join('.','Lifted_H2_subdomain') # change this with your path to the data folder
T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path '{T_path}' is correct...")
if not os.path.exists(T_path):
    raise ValueError("The path '{T_path}' does not exist in your system. "
                     "Change directory with the correct path to the data folder")

# Blastnet's datasets contain information about the shape of the field in the info.json file
import json
with open(os.path.join(directory,'info.json'), 'r') as file:
    info = json.load(file)
DNS_shape = info['global']['Nxyz']
    
# Now we have the shape and the path of the file, we can define the Scalar object:
T = ap.Scalar(shape=DNS_shape, path=T_path)
# Try to access the value of the temperature field:
T_values = T._3D #this will return a 3d array with the values of Temperature
print("Temperature value in the cell 10,10,10")
print(T_values[10,10,10])

"""
The object is now defined and we can access its values. Nothing special until now.
But let's check why using the Scalar object works so well with DNS data.
Now we are going to define a numpy array of the same size as the temperature
field, and we'll compare how much memory it takes in your system:
"""
T_numpy = np.random.rand(*DNS_shape)
import sys
T_size = sys.getsizeof(T)
T_numpy_size = sys.getsizeof(T_numpy)

print(f"\nSize of the numpy array:         {T_numpy_size} bytes"
      f"\nSize of the Scalar object:     {T_size} bytes")

"""
Showed the advantage of the object, in the following section the goal is to learn
how to use the method cut to cut this field.
To cut a different amount of cells in every direction x, y and z, we can use
the mode 'xyz'. This mode allows us to cut a different number of cells from
the 3 directions, but the number of cells cut, for example, in the x direction,
will be symmetric in the two directions of the x axis. If we cut 3 cells in
the x direction, the function will remove 3 cells at the beginning and 3 at the end.
"""
T_cut = T.cut(n_cut=[30, 0, 10], mode='xyz')
print(f'Initial shape:                               {T.shape}')
print(f'Scalar\'s shape after cutting with \'xyz\' mode:{T_cut.shape}')




# tests/conftest.py
import os
import matplotlib
import pytest


@pytest.fixture(scope="session", autouse=True)
def _mpl_headless():
    matplotlib.use("Agg")


@pytest.fixture(scope="session")
def apriori_test_cache_dir(tmp_path_factory):
    """
    Persistent cache for test datasets.

    Priority:
      1) APRIORI_TEST_DATA env var (if set)
      2) <repo_root>/.pytest_apriori_cache  (created automatically)
      3) pytest temp dir fallback (should rarely happen)
    """
    # 1) user override
    env = os.environ.get("APRIORI_TEST_DATA")
    if env:
        os.makedirs(env, exist_ok=True)
        return env

    # 2) repo-local persistent cache
    # pytest always runs with a working directory; typically repo root
    repo_cache = os.path.abspath(os.path.join(os.getcwd(), ".pytest_apriori_cache"))
    os.makedirs(repo_cache, exist_ok=True)
    return repo_cache


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 14:50:44 2024

@author: lorenzo piu
"""

import os
import aPriori as ap
from aPriori.DNS import Field
from aPriori import DNS
import json

# Uncomment the following line if you did not download the dataset yet
# ap.download(dataset='h2_lifted')

directory = os.path.join('.','Lifted_H2_subdomain') # change this with your path to the data folder

# Check the folder with the data exists in your system
T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path \'{T_path}\' is correct...")
if not os.path.exists(T_path):
    raise ValueError("The path '{T_path}' does not exist in your system. Check to have the correct path to your data folder in the code")
else:
    print("Folder path OK\n")

DNS_field = Field(directory)

cut_field_name = DNS_field.cut([40,50,10])

DNS_field_cut = Field(cut_field_name)

DNS_field.plot_z_midplane('U_X',
                          vmin=100,
                          vmax=280, 
                          transparent=False,
                          cbar_title=r'$U_x [m/s]$',
                          title='Uncut field',
                          dpi=500
                          )


DNS_field_cut.plot_z_midplane('U_X',
                              vmin=100,
                              vmax=280, 
                              transparent=False,
                              cbar_title=r'$U_x [m/s]$',
                              title='Cut field',
                              dpi=500
                              )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""derivatives.py: the module that handles derivatives (gradients and laplacians).

"""

__authors__ = "Lorenzo Piu"
__copyright__ = "Copyright (c) 2024-2026, Lorenzo Piu, Heinz Pitsch, and Alessandro Parente"
__credits__ = ["Aero-Thermo-Mechanics laboratories - Universite Libre de Bruxelles, Brussels, Belgium"
               "Institut für Technische Verbrennung (ITV) - RWTH Aachen University, Aachen, Germany"]
__license__ = "GPL 3.0"
__version__ = "1.3.0"
__maintainer__ = ["Lorenzo Piu"]
__email__ = ["lorenzo.piu@ulb.be"]
__status__ = "Production"


import numpy as np
from findiff import FinDiff
from .DNS import Scalar
from .DNS import Mesh


_DEFAULT_DERIVATIVES_ORDER = 4

def set_gradients_order(order: int) -> None:
    global _DEFAULT_DERIVATIVES_ORDER
    _DEFAULT_DERIVATIVES_ORDER = order

def print_gradients_order(order=None):
    if order is None:
        order = _DEFAULT_DERIVATIVES_ORDER
    print(f"Using order {order}")



def _apply_findiff_on_strided_grid(data, coords_1d, axis, filter_size, acc, deriv_order):
    """
    Internal helper function to compute derivatives on a potentially 
    strided (downsampled) grid using findiff.

    Parameters
    ----------
    data : numpy.ndarray
        The 3D array on which to compute the derivative.
    coords_1d : numpy.ndarray
        The 1D coordinates corresponding to the 'axis' of interest.
    axis : int
        The axis (0, 1, or 2) along which to compute the derivative.
    filter_size : int
        The stride/downsampling factor.
    acc : int
        Accuracy order for findiff.
    deriv_order : int
        The order of the derivative (1 for gradient, 2 for Laplacian).

    Returns
    -------
    result : numpy.ndarray
        The computed derivative with the same shape as 'data'.
    """
    result = np.zeros_like(data)

    # Loop over the stride offsets
    for i in range(filter_size):
        # 1. Create the slice object for the 3D data
        # We want equivalent of data[i::filter_size, :, :] if axis=0
        slices = [slice(None)] * 3
        slices[axis] = slice(i, None, filter_size)
        tuple_slices = tuple(slices)

        # 2. Extract the strided data
        sub_field = data[tuple_slices]

        # 3. Extract the strided coordinates for this specific axis
        sub_coords = coords_1d[i::filter_size]

        # 4. Check if we have enough points for the requested accuracy
        # Findiff generally requires roughly 'acc' points. 
        if len(sub_coords) <= acc:
             print(f"Warning: Grid too coarse (N={len(sub_coords)}) for requested accuracy ({acc}) on stride {i}.")

        # 5. Define and apply the FinDiff operator
        # FinDiff(axis, spacing/coords, derivative_order, accuracy)
        fd = FinDiff(axis, sub_coords, deriv_order, acc=acc)
        
        # 6. Compute and place back into the result array
        result[tuple_slices] = fd(sub_field)

    return result


def gradient_x(F, mesh, filter_size=1, acc=8, reduce_acc=False):
    '''
    Description
    -----------
    Computes the x-component of the gradient using findiff with specified accuracy.
    Uses striding (filter_size) to compute derivatives on downsampled grids.

    Parameters
    ----------
    F : Scalar object
        The field to differentiate.
    mesh : Mesh object
        The mesh object containing coordinates (X1D).
    filter_size : int, optional
        The stride used to filter the field (default is 1).
    acc : int, optional
        The accuracy order for the finite difference scheme (default is 2).

    Returns
    -------
    grad_x : numpy array
        The x component of the gradient.
    '''
    # 1. Type Checking
    # Note: Assuming Scalar/Mesh names are available in the namespace. 
    # If not imported, use F.__class__.__name__ == 'Scalar' or similar.
    if type(F).__name__ != 'Scalar':
        raise TypeError("F must be an object of the class Scalar")
    if type(mesh).__name__ != 'Mesh':
        raise TypeError("mesh must be an object of the class Mesh")
    if not isinstance(filter_size, int):
        raise TypeError("filter_size must be an integer")
    if not isinstance(acc, int):
        raise TypeError("acc must be an integer")
    
    if reduce_acc: # Switch to decrease gradients accuracy (used for example on LES grids)
        acc=2

    # 2. Call Internal Helper
    return _apply_findiff_on_strided_grid(
        data=F._3d,
        coords_1d=mesh.X1D,
        axis=0,
        filter_size=filter_size,
        acc=acc,
        deriv_order=1
    )


def gradient_y(F, mesh, filter_size=1, acc=8, reduce_acc=False):
    '''
    Description
    -----------
    Computes the y-component of the gradient using findiff.
    '''
    if type(F).__name__ != 'Scalar':
        raise TypeError("F must be an object of the class Scalar")
    if type(mesh).__name__ != 'Mesh':
        raise TypeError("mesh must be an object of the class Mesh")
    if not isinstance(acc, int):
        raise TypeError("acc must be an integer")
    
    if reduce_acc:
        acc = 2

    return _apply_findiff_on_strided_grid(
        data=F._3d,
        coords_1d=mesh.Y1D,
        axis=1,
        filter_size=filter_size,
        acc=acc,
        deriv_order=1
    )


def gradient_z(F, mesh, filter_size=1, acc=8, reduce_acc=False):
    '''
    Description
    -----------
    Computes the z-component of the gradient using findiff.
    '''
    if type(F).__name__ != 'Scalar':
        raise TypeError("F must be an object of the class Scalar")
    if type(mesh).__name__ != 'Mesh':
        raise TypeError("mesh must be an object of the class Mesh")
    if not isinstance(acc, int):
        raise TypeError("acc must be an integer")
    
    if reduce_acc:
        acc = 2

    return _apply_findiff_on_strided_grid(
        data=F._3d,
        coords_1d=mesh.Z1D,
        axis=2,
        filter_size=filter_size,
        acc=acc,
        deriv_order=1
    )


def laplacian(F, mesh, filter_size=1, acc=4, reduce_acc=False):
    '''
    Description
    -----------
    Computes the Laplacian (d2/dx2 + d2/dy2 + d2/dz2) of the field.
    
    Parameters
    ----------
    F : Scalar object
    mesh : Mesh object
    filter_size : int, optional
    acc : int, optional

    Returns
    -------
    lap : numpy array
        The Laplacian of the input field.
    '''
    if type(F).__name__ != 'Scalar':
        raise TypeError("F must be an object of the class Scalar")
    if type(mesh).__name__ != 'Mesh':
        raise TypeError("mesh must be an object of the class Mesh")
    
    if reduce_acc:
        acc = 2

    # Compute second derivatives for each axis
    laplacian_value  = _apply_findiff_on_strided_grid(F._3d, mesh.X1D, 0, filter_size, acc, deriv_order=2)
    laplacian_value += _apply_findiff_on_strided_grid(F._3d, mesh.Y1D, 1, filter_size, acc, deriv_order=2)
    laplacian_value += _apply_findiff_on_strided_grid(F._3d, mesh.Z1D, 2, filter_size, acc, deriv_order=2)

    return laplacian_value


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 14:10:20 2024

@author: lorenzo piu
"""

import os
from aPriori.DNS import Field
from aPriori import DNS
import aPriori as ap
import json

# Comment the following line if you already downloaded the dataset
ap.download(dataset='h2_lifted')

# Change this with your path to the data folder if necessary
directory = os.path.join('.','Lifted_H2_subdomain') 

# Check the folder with the data exists in your system
T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path \'{T_path}\' is correct...")
if not os.path.exists(T_path):
    raise ValueError("The path '{T_path}' does not exist in your system. Check to have the correct path to your data folder in the code")
else:
    print("Folder path OK\n")

# Blastnet's data contain information about the shape of the field in the info.json file
import json
with open(os.path.join(directory,'info.json'), 'r') as file:
    info = json.load(file)
DNS_shape = info['global']['Nxyz']
    
DNS_field = Field(directory)

DNS_field.T

DNS_field.U_Y._3D

filt_YO2 = DNS.filter_3D(DNS_field.YO2._3D, 8)

DNS_field.plot_z_midplane('YH2O2')





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Reference plotting parameters and styles that will be used in the plot_utils module.
"""
__authors__ = "Lorenzo Piu"
__copyright__ = "Copyright (c) 2024-2026, Lorenzo Piu, Heinz Pitsch, and Alessandro Parente"
__credits__ = ["Aero-Thermo-Mechanics laboratories - Universite Libre de Bruxelles, Brussels, Belgium"
               "Institut für Technische Verbrennung (ITV) - RWTH Aachen University, Aachen, Germany"]
__license__ = "GPL 3.0"
__version__ = "1.3.0"
__maintainer__ = ["Lorenzo Piu"]
__email__ = ["lorenzo.piu@ulb.be"]
__status__ = "Production"


class ParityPlot():
    fig_size      = [10, 8]
    dpi           = 400
    fontsize      = 48
    border_width  = 2
    
class ContourPlot():
    fig_size     = 4 # Will be adjusted based on the aspect ratio
    dpi          = 800
    fontsize     = 32
    
class CondMeanPlot():
    fig_size     = [8, 6]
    dpi          = 400
    fontsize     = 48
    border_width = 2

class ScatterPlot():
    fig_size     = [8, 6]
    dpi          = 400
    fontsize     = 48
    aspect_ratio = 1.2
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on February 24 16:41:09 2026

@author: lorenzo piu
"""

import os
import aPriori as ap
from aPriori import DNS
import json

# comment the following line if you don't want to download the example dataset
# ap.download(dataset='csp')

directory = os.path.join('.', 'Lifted_H2_subdomain_csp') # change this with your path to the data folder
figures_folder = 'figures_csp'

dns_field = ap.Field(directory)

dns_field.compute_csp(
    API_species=['H2O', 'OH'], 
    n_chunks=100,
    )

os.makedirs(figures_folder, exist_ok=True)

dns_field.plot_api(
    api_type='TSR', 
    n=10, 
    n_cols=5, 
    cmap='RdGy_r', 
    vmin=-1, vmax=1,
    save_path=os.path.join(figures_folder, 'api_tsr.png')
    )

dns_field.plot_api(
    api_type='OH', 
    n=10, 
    n_cols=5, 
    cmap='RdGy_r', 
    vmin=-1, vmax=1,
    save_path=os.path.join(figures_folder, 'api_oh.png')
    )

dns_field.plot_z_midplane(
    'TSR_DNS',
    colormap='gist_heat',
    transpose=True,
    remove_title=True,
    cbar_title='TSR $[s^{-1}]$',
    vmax=0,
    transparent=False,
    save_path=os.path.join(figures_folder, 'TSR.png')
    )

dns_field.plot_z_midplane(
    'TSR_DNS',
    colormap='gist_heat_r',
    transpose=True,
    remove_title=True,
    cbar_title='TSR $[s^{-1}]$',
    vmin=0,
    transparent=False,
    remove_y=True,
    save_path=os.path.join(figures_folder, 'TSR_pos.png')
    )

dns_field.plot_z_midplane(
    'YOH',
    colormap='gist_heat_r',
    transpose=True,
    remove_title=True,
    cbar_title=r'$Y_{OH}$ $[-]$',
    save_path=os.path.join(figures_folder, 'YOH.png')
    )

dns_field.plot_z_midplane(
    'T',
    colormap='gist_heat_r',
    transpose=True,
    remove_title=True,
    remove_y=True,
    cbar_title=r'$T$ $[K]$',
    save_path=os.path.join(figures_folder, 'T.png')
    )




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 14:50:44 2024

@author: lorenzo piu
"""

import os
import aPriori as ap
from aPriori import DNS
import json

# Uncomment the following line if you did not download the dataset yet
# ap.download(dataset='h2_lifted')

directory = os.path.join('.', 'Lifted_H2_subdomain') # change this with your path to the data folder

T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path \'{T_path}\' is correct...")
if not os.path.exists(T_path):
    print(f"The path '{T_path}' does not exist in your system. Downloading the dataset from Github...")
    ap.download(dataset='h2_lifted')
    
DNS_field = ap.Field(directory)

filter_size = 32

# Favre filter with Gaussian filter kernel
field_filt_name = DNS_field.filter_favre(filter_size, filter_type='Gauss')
DNS_field_filt_gauss = ap.Field(field_filt_name)

# Favre filter with box filter kernel
# The box filter can take a long time as the convolution operation is not
# optimized. If you have any suggestion on how to improve it, please
# open an issue on the Github page!
DNS_field_filt_box = ap.Field(DNS_field.filter_favre(filter_size, filter_type='Box'))

# Plot the different filtering results
DNS_field.plot_z_midplane('T',
                          title='DNS',
                          cbar_title='T [K]',
                          colormap='inferno',
                          vmin=500,
                          vmax=2300,
                          transparent=False,
                          transpose=True,
                          remove_cbar=True
                          )

DNS_field_filt_box.plot_z_midplane('T',
                          title='$\Delta = 32$, Box',
                          cbar_title='T [K]',
                          colormap='inferno',
                          vmin=500,
                          vmax=2300,
                          transparent=False,
                          transpose=True,
                          remove_cbar=True,
                          remove_y=True
                          )

DNS_field_filt_gauss.plot_z_midplane('T',
                          title='$\Delta = 32$, Gauss',
                          cbar_title='T [K]',
                          colormap='inferno',
                          vmin=500,
                          vmax=2300,
                          transparent=False,
                          transpose=True,
                          remove_y=True
                          )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 14:50:44 2024

@author: lorenzo piu
"""

import aPriori as ap

# Download the dataset
ap.download(dataset='h2_lifted')

# Initialize 3D DNS field
field_DNS = ap.Field('Lifted_H2_subdomain')

#----------------------------Visualize the dataset-----------------------------

# Plot Temperature on the xy midplane (transposed as yx plane)
field_DNS.plot_z_midplane('T',                 # plots the Temperature
                          levels=[1400, 2000], # isocontours at 1400 and 2000
                          vmin=1400,           # minimum temperature to plot
                          title='T [K]',       # figure title
                          linewidth=2,         # isocontour lines thickness
                          transpose=True,      # inverts x and y axes
                          x_name='y [mm]',     # x axis label
                          y_name='x [mm]')     # y axis label
# Plot Temperature on the xz midplane (transposed as zx plane)
field_DNS.plot_y_midplane('T', 
                          levels=[1400, 2000], 
                          vmin=1400, 
                          title='T [K]', 
                          linewidth=2,
                          transpose=True, 
                          x_name='z [mm]', 
                          y_name='x [mm]')
# Plot Temperature on the yz midplane
field_DNS.plot_x_midplane('T', levels=[1400, 2000], vmin=1400, 
                          title='T [K]', linewidth=2)

#--------------------------Compute DNS reaction rates--------------------------
field_DNS.compute_reaction_rates()

# Plot OH mass fraction on the transposed xy midplane
field_DNS.plot_z_midplane('YOH', title=r'$Y_{OH}$ $[-]$', colormap='inferno',
                          transpose=True, x_name='z [mm]', y_name='x [mm]')
# Plot reaction rates
field_DNS.plot_z_midplane('ROH_DNS', 
                          title=r'$\dot{\omega}_{OH}$ $[kg/m^3/s]$', 
                          colormap='inferno',
                          transpose=True, x_name='z [mm]', y_name='x [mm]')
field_DNS.plot_z_midplane('RH2O_DNS', 
                          title=r'$\dot{\omega}_{H2O}$ $[kg/m^3/s]$', 
                          colormap='inferno',
                          transpose=True, x_name='z [mm]', y_name='x [mm]')

#--------------------------Compute derived variables---------------------------
# compute kinetic energy
field_DNS.compute_kinetic_energy()

# Compute mixture fraction
field_DNS.ox = 'O2'     # Defines the species to consider as oxydizer
field_DNS.fuel = 'H2'   # Defines the species to consider as fuel
Y_ox_2=0.233  # Oxygen mass fraction in the oxydizer stream (air)
Y_f_1=0.65*2/(0.65*2+0.35*28) # Hydrogen mass fraction in the fuel stream
# (the fuel stream is composed by X_H2=0.65 and X_N2=0.35)

field_DNS.compute_mixture_fraction(Y_ox_2=Y_ox_2, Y_f_1=Y_f_1, s=2)

# Scatter plot variables as functions of the mixture fraction Z
field_DNS.scatter_Z('T', # the variable to plot on the y axis
                    c=field_DNS.YOH.value, # set color of the points
                    y_name='T [K]', 
                    cbar_title=r'$Y_{OH}$'
                    )

field_DNS.scatter_Z('ROH_DNS',
                    c=field_DNS.HRR_DNS.value, 
                    y_name=r'$\dot{\omega}_{OH}$ $[kg/m^3/s]$', 
                    cbar_title=r'$\dot{Q}_{DNS}$'
                    )

#-------------------------------Filter DNS field-------------------------------
# perform favre filtering (high density gradients)
# the output of the function is a string with the new folder's name, f_string
f_string = field_DNS.filter_favre(filter_size=16, # filter amplitude
                                        filter_type='Gauss') # 'Gauss' or 'Box'

# The string with the folder's name is now used to initialize the filered field
field_filtered = ap.Field(f_string)

# Visualize the effect of filtering on the Heat Release Rate
field_DNS.plot_z_midplane('HRR_DNS',
                          title=r'$\dot{Q}_{DNS}$ $[W/m^3]$', 
                          colormap='inferno',
                          vmax=8*1e9,
                          transpose=True, x_name='z [mm]', y_name='x [mm]',
                          remove_cbar=True)

field_filtered.plot_z_midplane('HRR_DNS',
                          title=r'$\overline{\dot{Q}_{DNS}}$ $[W/m^3]$', 
                          colormap='inferno',
                          vmax=8*1e9,
                          transpose=True, x_name='z [mm]', y_name='x [mm]',
                          remove_y=True)

#-------------------------Compute reaction rates (LFR)-------------------------
# Computing reaction rates directly from the filtered field (LFR approximation)
field_filtered.compute_reaction_rates()

# Compare visually the results 
field_filtered.plot_z_midplane('RH2_DNS',
                          title=r'$\overline{\dot{\omega}}_{H2}^{DNS}$ $[kg/(m^3\cdot s)]$', 
                          vmax=-20,
                          vmin=field_filtered.RH2_LFR.z_midplane.min(),
                          levels=[-300, -50, -20],
                          labels=True,
                          colormap='inferno',
                          transpose=True, x_name='z [mm]', y_name='x [mm]',
                          remove_cbar=True)

# Compare visually the results in the z midplane
field_filtered.plot_z_midplane('RH2_LFR',
                          title=r'$\overline{\dot{\omega}}_{H2}^{LFR}$ $[kg/(m^3\cdot s)]$', 
                          vmax=-20,
                          vmin=field_filtered.RH2_LFR.z_midplane.min(),
                          levels=[-300, -50, -20],
                          labels=True,
                          colormap='inferno',
                          transpose=True, x_name='z [mm]', y_name='x [mm]',
                          remove_y=True)

# Compare the heat release rate results with a parity plot
f = ap.parity_plot(field_filtered.HRR_DNS.value,  # x
                   field_filtered.HRR_LFR.value,  # y
                   density=True,                  # KDE coloured
                   x_name=r'$\overline{\dot{\omega}}_{H2}^{DNS}$',
                   y_name=r'$\overline{\dot{\omega}}_{H2}^{LFR}$',
                   cmin=0,
                   RMSE=False,
                   NRMSE=True,
                   ticks=[0, 0.8e10, 1.6e10]
                   )

#------------------------Compute reaction rates (PaSR)-------------------------
import sys
# WARNING: this computation step can take up to 1 hour
response = input("Computing the batch reactor reaction rates can take up to 20" 
                 "minutes.\n"
                 "Do you want to continue? ([yes]/no): ").strip().lower()
if response == 'yes':
    print("\nContinuing...")
else:
    print("\nExiting...")
    sys.exit()
    
# Compute the strain rate, and the residual k and epsilon
field_filtered.compute_strain_rate(save_tensor=False)
field_filtered.compute_residual_kinetic_energy()
field_filtered.compute_residual_dissipation_rate()
field_filtered.compute_mixing_timescale()
field_filtered.compute_chemical_timescale()

field_filtered.compute_reaction_rates_batch()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 12:00:20 2024

@author: lorenzo piu
"""

import aPriori.DNS as DNS
from aPriori.DNS import Scalar
import numpy as np

shape = [4, 2, 4]
array = np.ones(shape)
array[2:3, :, 2:3] = 2
print(array)

filter_size = 2
filtered_field = DNS.filter_3D(array, filter_size, favre=False, filter_type='Gauss')
print(filtered_field)

import os
directory = os.path.join('..','data','Lifted_H2_subdomain') # change this with your path to the data folder
T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path \'{T_path}\' is correct...")
if not os.path.exists(T_path):
    raise ValueError("The path '{T_path}' does not exist in your system. Check to have the correct path to your data folder in the code")

# Blastnet's data contain information about the shape of the field in the info.json file
import json
with open(os.path.join(directory,'info.json'), 'r') as file:
    info = json.load(file)
DNS_shape = info['global']['Nxyz']
    
# Now we have the shape and the path of the file, we can define the Scalar object:
T = Scalar(shape=DNS_shape, path=T_path)
# Try to access the value of the temperature field:
print("Temperature value in the cells 5:8, 5:8, 5:8")
print(T._3D[5:8, 5:8, 5:8])

filter_size = 20
T_filt_gauss = DNS.filter_3D(T._3D, filter_size, favre=False, filter_type='Gauss')
print("Filtered temperature values in the cells 5:8, 5:8, 5:8")
print(T_filt_gauss[5:8, 5:8, 5:8])

T_filt_box = DNS.filter_3D(T._3D, filter_size, favre=False, filter_type='Box')
print("Filtered temperature values in the cells 5:8, 5:8, 5:8")
print(T_filt_box[5:8, 5:8, 5:8])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Plotting Utilities
"""

__authors__ = "Lorenzo Piu"
__copyright__ = "Copyright (c) 2024-2026, Lorenzo Piu, Heinz Pitsch, and Alessandro Parente"
__credits__ = ["Aero-Thermo-Mechanics laboratories - Universite Libre de Bruxelles, Brussels, Belgium"
               "Institut für Technische Verbrennung (ITV) - RWTH Aachen University, Aachen, Germany"]
__license__ = "GPL 3.0"
__version__ = "1.3.0"
__maintainer__ = ["Lorenzo Piu"]
__email__ = ["lorenzo.piu@ulb.be"]
__status__ = "Production"

import matplotlib.pyplot as plt
import matplotlib
import numpy as np
from scipy.stats import gaussian_kde
from matplotlib import ticker
from matplotlib.colors import LogNorm

from ._styles import ParityPlot, ContourPlot, CondMeanPlot, ScatterPlot
from ._utils import extract_random_elements, reorder_arrays, process_arrays

plt.rcParams.update({
    'font.size': 18,           # Default font size
    'figure.figsize': (10, 6), # Default figure size (width, height) in inches
    # 'axes.titlesize': 18,      # Title font size
    # 'axes.labelsize': 14,      # Axis label font size
    # 'xtick.labelsize': 12,     # X-axis tick label font size
    # 'ytick.labelsize': 12,     # Y-axis tick label font size
    # 'legend.fontsize': 12,     # Legend font size
    'lines.linewidth': 2,      # Line width
    # 'lines.markersize': 6,     # Marker size
    'axes.linewidth': 1.5,     # Axes border width
    'grid.linewidth': 1,       # Grid line width
    # 'xtick.major.width': 1,    # Major tick width on x-axis
    # 'ytick.major.width': 1,    # Major tick width on y-axis
    # 'xtick.minor.width': 0.8,  # Minor tick width on x-axis
    # 'ytick.minor.width': 0.8   # Minor tick width on y-axis
})

def parity_plot(x, 
                y, 
                rel_error=0.2, 
                colormap='viridis', 
                x_name='x',
                y_name='y',
                title=None,
                density=False, 
                bins=200,
                cmin=1,
                c=None, 
                linestyle='--',
                linewidth=2.5,
                vmin=None,
                vmax=None,
                cbar_title=None,
                marker='.',
                s=1, 
                alpha=1,
                R2=True,
                RMSE=True,
                NRMSE=False,
                ticks=None,
                limits=None,
                save=False,
                show=True,
                max_dim=100000,
                remove_cbar=False,
                remove_x=False,
                remove_y=False
                ):
    
    # Flatten the vectors because they will be stack afterwards
    x = x.flatten()
    y = y.flatten()
    
    # Find the limits of the domain
    min_x = np.minimum(np.min(x), np.min(y))
    max_x = np.maximum(np.max(x), np.max(y))
    
    if limits is None:
        limits = [min_x, max_x]
    else:
        min_x = limits[0]
        max_x = limits[1]
        
    # Compute the error metrics if required:
    if R2:
        R2 = 1 - np.sum((x-y)**2) / np.sum((x-np.average(x))**2)
    if RMSE:
        RMSE = np.sqrt(np.mean((x - y) ** 2))
    if NRMSE:
        NRMSE = np.sqrt(np.mean((x - y) ** 2))/np.max(x)
    
    if density:
        cbar_title='PDF'

    # PLOT
    fig = plt.figure(figsize=ParityPlot.fig_size, 
                     dpi=ParityPlot.dpi)
    
    if not density:
        if c is None:
            c='blue'
        plt.scatter(x, y, marker=marker, c=c, s=s, 
                    cmap=colormap, alpha=alpha, vmin=vmin, vmax=vmax)
    else:
        plt.hist2d(x, y, 
                   bins=bins, 
                   cmin=cmin, 
                   density=True, 
                   cmap=colormap,
                   norm=LogNorm()
                   )
    
    if rel_error is not None:
        plt.fill_between(limits, 
                         [min_x*(1-rel_error), max_x*(1-rel_error)], 
                         [min_x*(1+rel_error), max_x*(1+rel_error)], 
                         color='grey', alpha=0.4)
        
    if (not isinstance(c, str)) or (c is None): # if c is a string means it was initialized as a single color, so no need for colormap
        cbar = plt.colorbar(shrink=.9, aspect=15, fraction=.1,pad=.05)
        cbar.ax.tick_params(labelsize=ParityPlot.fontsize*3//4)
        if cbar_title is not None:
            cbar.set_label(cbar_title, rotation=90, fontsize=ParityPlot.fontsize)
    plt.xlabel(x_name, fontsize=ParityPlot.fontsize)
    plt.ylabel(y_name, fontsize=ParityPlot.fontsize)
    if title is not None:
        plt.title(title, fontsize=ParityPlot.fontsize)
    plt.xlim(limits)
    plt.ylim(limits)
    plt.xticks(fontsize=ParityPlot.fontsize*4//5)
    plt.yticks(fontsize=ParityPlot.fontsize*4//5)
    if ticks is not None:
        # if isinstance(ticks, int):
        # These lines create n equispaced ticks
        #     xmin    = plt.gca().get_xlim()[0]
        #     xmax    = plt.gca().get_xlim()[1]
        #     ticks = np.linspace(xmin, xmax, ticks)
        plt.xticks(ticks=ticks)
        plt.yticks(ticks=ticks)
    plt.plot(limits, limits, 'k', linestyle=linestyle, linewidth=linewidth)
    
    # Write the error metrics if required
    lines = []
    if NRMSE:  lines.append(rf'$\mathrm{{NRMSE}} = {NRMSE:.2e}$')
    if R2:   lines.append(rf'$R^2 = {R2:.4f}$')
    if RMSE:  lines.append(rf'$\mathrm{{RMSE}} = {RMSE:.4f}$')
    annotation = "\n".join(lines)
    if annotation:
        plt.text(0.02, 0.98, annotation, 
                 transform=plt.gca().transAxes, 
                 fontsize=ParityPlot().fontsize*0.65,
                 va='top',
                 ha='left'
                 )
    # Adjust borders tickness
    for spine in plt.gca().spines.values():
        spine.set_linewidth(ParityPlot.border_width )
    
    plt.gca().set_aspect('equal', adjustable='box')
    if remove_cbar:
        cbar.remove()
    if remove_x:
        plt.xlabel('')
        plt.xticks([])
    if remove_y:
        plt.ylabel('')
        plt.yticks([])
    figure = plt.gcf()
    axes = plt.gca()
    
    if save:
        if title is None:
            title=f'Parity_plot_{x_name}_{y_name}'
        plt.savefig(title, dpi=ParityPlot.dpi, bbox_inches = "tight")
    
    if show:
        plt.show()
        
    # # Close the figure to prevent it from displaying when returned
    # plt.close(figure)
    
    return figure, axes
    

def contour_plot(X, 
                 Y, 
                 Z,
                 log=False,
                 colormap='viridis', 
                 cbar_title=None,
                 cbar_shrink=0.7,
                 levels=None,
                 color='black',
                 labels=False,
                 linestyle='-',
                 linecolor='black',
                 linewidth=1,
                 x_ticks=None,
                 y_ticks=None,
                 x_lim=None,
                 y_lim=None,
                 vmin=None,
                 vmax=None,
                 transparent=True,
                 title=None,
                 title_pad=20,
                 x_name='x',
                 y_name='y',
                 remove_cbar=False,
                 remove_x=False,
                 remove_y=False,
                 remove_title=False,
                 dpi=None,
                 transpose=False,
                 save=False,
                 save_path=None,
                 show=True
                 ):
    
    if dpi is None:
        dpi = ContourPlot.dpi
    
    Z = Z.copy() # so it won't be modified inside the funciton
    X = X.copy()
    Y = Y.copy()
    if log:
        Z = np.log10(Z)
    
    if transpose:
        temp = X
        X = Y
        Y = temp
    
    ratio = np.max(X)-np.min(X)/X.shape[1]
    figsize = [ContourPlot.fig_size, ratio*ContourPlot.fig_size]
    fontsize=ContourPlot.fontsize
    
    # Create the contour plot
    plt.figure(dpi=ContourPlot.dpi)
    # if transparent, plot the points and then cut the vectors
    if transparent:
        # plt.pcolormesh(X, Y, Z, color='white')
        Z_full = Z.copy()
        if (vmin is not None) and (vmax is not None):
            Z[(Z<vmin) | (Z>vmax)] = None
        else:
            if (vmin is not None):
                if (np.min(Z) < vmin):
                    Z[Z<vmin] = None
            if (vmax is not None):
                if (np.max(Z) > vmax):
                    Z[Z>vmax] = None
        
    # Plot filled contours
    contourf = plt.pcolormesh(X, Y, Z, cmap=colormap, vmin=vmin, vmax=vmax)  # Filled contours
    
    if levels is not None:
        # Plot iso-lines with all lines in black
        contour = plt.contour (X, Y, Z_full, 
                              levels=levels, 
                              colors=linecolor, 
                              linestyles=linestyle, 
                              linewidths=linewidth)  # All lines in black
    
    # Add a color bar to the plot
    cbar = plt.colorbar(contourf, shrink=cbar_shrink, aspect=15, fraction=.1,pad=.05)
    cbar.ax.tick_params(labelsize=ContourPlot.fontsize*3//4)
    cbar.ax.tick_params(labelsize=ContourPlot.fontsize*3//4)
    if log:
        # Get colorbar limits
        vmin, vmax = cbar.mappable.get_clim()
        # Get existing ticks, and keep only those within the colorbar range
        ticks = [tick for tick in cbar.get_ticks() if vmin <= tick <= vmax]
        # Format ticks as 10^n (using integer or float depending on tick)
        tick_labels = [f'$10^{{{int(t)}}}$' if t == int(t) else f'$10^{{{t:.1f}}}$' for t in ticks]
        # Set ticks and labels manually
        cbar.set_ticks(ticks)
        cbar.set_ticklabels(tick_labels)
    if cbar_title is not None:
        # cbar.set_label(cbar_title, rotation=270, fontsize=ContourPlot.fontsize )
        
        # bbox = cbar.ax.get_position()  # [x0, y0, width, height] in figure coords
        # x = bbox.x0 #+ 0.5*bbox.width
        # y = bbox.y1 + 0.10              # 0.01 pushes it a little above the bar
        # plt.gca().text(x, y, cbar_title, ha='center', va='bottom', fontsize=fontsize)
        cbar.ax.text(-0.05, 1.08, cbar_title,
             ha='left', va='bottom',
             transform=cbar.ax.transAxes,
             fontsize=fontsize*1.1)
    
    # Add labels and title
    plt.xlabel(x_name, fontsize=fontsize)
    plt.ylabel(y_name, fontsize=fontsize)
    if title is not None:
        plt.title(title, fontsize=fontsize, pad=title_pad)
    
    # x and y limits
    if x_lim is not None:
        plt.xlim(x_lim)
    if y_lim is not None:
        plt.ylim(y_lim)
    
    # Handle ticks
    if x_ticks is not None:
        if isinstance(x_ticks, int):
            xmin    = plt.gca().get_xlim()[0]
            xmax    = plt.gca().get_xlim()[1]
            x_ticks = np.linspace(xmin, xmax, x_ticks)
        plt.xticks(ticks=x_ticks)
    if y_ticks is not None:
        if isinstance(y_ticks, int):
            ymin    = plt.gca().get_ylim()[0]
            ymax    = plt.gca().get_ylim()[1]
            y_ticks = np.linspace(ymin, ymax, y_ticks)
        plt.yticks(ticks=y_ticks)
    plt.xticks(fontsize=fontsize*3//4)
    plt.yticks(fontsize=fontsize*3//4)
    if labels:
        # Optionally, add labels to contour lines
        plt.clabel(contour, inline=True, fontsize=fontsize//4, fmt='%1.1f')
        
    plt.gca().set_aspect('equal', adjustable='box')
    if remove_cbar:
        cbar.remove()
    if remove_x:
        plt.xlabel('')
        plt.xticks([])
    if remove_y:
        plt.ylabel('')
        plt.yticks([])
    if remove_title:
        plt.title('')
    
    figure = plt.gcf()
    axes = plt.gca()
    
    if save or (save_path is not None):
        if save_path is None:
            if title is None:
                title=f'Contour_plot'
                if cbar_title is not None:
                    title = title+'_'+cbar_title
            else:
                title = 'contour_plot' + title
        else:
            title = save_path
        plt.savefig(title, dpi=dpi, bbox_inches = "tight")
    
    if show:
        plt.show()
        
    # Close the figure to prevent it from displaying when returned
    plt.close(figure)
    
    return figure, axes

def cond_mean_plot(x, y, num_bins=30, 
                   log=False, 
                   minmax=True,
                   variance=False,
                   background=False,
                   background1=False,
                   colors=None,
                   markers=None,
                   x_name=None,
                   y_name=None,
                   title=None,
                   x_ticks=None,
                   y_ticks=None,
                   remove_markers=True,
                   remove_x=False,
                   remove_y=False,
                   save=False,
                   show=True
                   ):
    """
    Plots the conditional mean of y over x using numpy.

    Parameters:
    x (array-like): Independent variable.
    y (array-like): Dependent variable.
    num_bins (int): Number of bins to divide x into.
    """
    
    # check that x and y have the same dimension.
    # if y contains more than 1 vector, extracts m, the number of vectors it contains.
    # it also makes y a list, so that it is iterable
    x, y, m = process_arrays(x, y)
    
    # style
    lines = None
    if colors is None:
        colors = generate_colors(m)
    if markers is None:
        markers = generate_markers(m)
    elif markers is False:
        markers = ['' for i in range(m)]
        lines   = generate_lines(m)
    if lines is None:
        lines = ['-' for i in range(m)]
        
    fontsize = CondMeanPlot.fontsize
    dpi      = CondMeanPlot.dpi
    fig_size = CondMeanPlot.fig_size
    
    # initialize figure
    plt.figure(figsize=fig_size, dpi=dpi)
    
    # iterate for the number of y
    for i in range(m):
        x_b,y_b,y_max,y_min,y_var = bins(x, y[i], n=num_bins, log=log)
        color = colors[i]
        # Plotting
        plt.plot(x_b, y_b, marker=markers[i], linestyle=lines[i], c=color)
        if minmax:
            plt.fill_between(x_b, y_max, y_min, alpha=0.45, color=color)
        if variance:
            plt.fill_between(x_b, y_b+y_var, y_b-y_var, alpha=0.55, color=color)
        if background:
            plt.scatter(x, y[i], marker='.', alpha=0.3, c=color)
        if background1 and i==0:
            plt.scatter(x, y[i], marker='.', alpha=0.6, c='#c2c2c2') #c='#d3d3d3')
    

    plt.grid(True, linestyle='--', alpha=0.7)
    # Add labels
    plt.xlabel(x_name, fontsize=fontsize)
    plt.ylabel(y_name, fontsize=fontsize)
    
    # Adjust borders tickness
    for spine in plt.gca().spines.values():
        spine.set_linewidth(ParityPlot.border_width )
    
    # Handle ticks
    if x_ticks is not None:
        if isinstance(x_ticks, int):
            xmin    = plt.gca().get_xlim()[0]
            xmax    = plt.gca().get_xlim()[1]
            x_ticks = np.linspace(xmin, xmax, x_ticks)
        plt.xticks(ticks=x_ticks)
    if y_ticks is not None:
        if isinstance(y_ticks, int):
            ymin    = plt.gca().get_ylim()[0]
            ymax    = plt.gca().get_ylim()[1]
            y_ticks = np.linspace(ymin, ymax, y_ticks)
        plt.yticks(ticks=y_ticks)
    plt.xticks(fontsize=fontsize*3//4)
    plt.yticks(fontsize=fontsize*3//4)

    figure = plt.gcf()
    if remove_x:
        plt.xlabel('')
        plt.xticks([])
    if remove_y:
        plt.xlabel('')
        plt.xticks([])
    
    figure = plt.gcf()
    
    if save:
        if title is None:
            title=f'Contour_plot'
        else:
            title = 'Conditional_mean_' + title
        plt.savefig(title, dpi=dpi, bbox_inches = "tight")
    
    if show:
        plt.show()
        
    # Close the figure to prevent it from displaying when returned
    plt.close(figure)
    
    return figure

def scatter(x, y,  logx=False, 
                   logy=False,
                   c=None,
                   s=1,
                   alpha=1,
                   max_dim=2000000,
                   colormap='viridis',
                   cbar_title=None,
                   vmin=None,
                   vmax=None,
                   marker='.',
                   avg=True,
                   avg_label=None,
                   num_bins=100, 
                   linestyle='--',
                   linewidth=2,
                   linecolor='b',
                   Z_st=None,
                   x_name=None,
                   y_name=None,
                   title=None,
                   x_ticks=None,
                   y_ticks=None,
                   remove_markers=True,
                   remove_x=False,
                   remove_y=False,
                   save=False,
                   show=True
                   ):
    
    # style
    fontsize = ScatterPlot.fontsize
    dpi      = ScatterPlot.dpi
    fig_size = ScatterPlot.fig_size
    
    if (c is not None) and (len(x) > max_dim):
        print(f"\nScatter Plot:\nThe vector size is too large to display"
              f" the scatter plot. A random sample of {max_dim} elements "
              "will be extracted from the vector for the plot. You can "
              "change this limit with the parameter max_dim")
        x = extract_random_elements(x, max_dim, seed=42)
        y = extract_random_elements(y, max_dim, seed=42)
        c = extract_random_elements(c, max_dim, seed=42)
    
    # initialize figure
    plt.figure(figsize=fig_size, dpi=dpi)
    
    plt.scatter(x, y, s=s, c=c, marker=marker, alpha=alpha)
    if (not isinstance(c, str)) and (c is not None): # if c is a string means it was initialized as a single color, so no need for colormap
        cbar = plt.colorbar(shrink=.9, aspect=15, fraction=.1,pad=.05)
        cbar.ax.tick_params(labelsize=fontsize*3//4)
        if cbar_title is not None:
            cbar.set_label(cbar_title, rotation=90, fontsize=ParityPlot.fontsize)

    if avg is True:
        x_b,y_b,y_max,y_min,y_var = bins(x, y, n=num_bins, log=logx)
        plt.plot(x_b, y_b, linewidth=linewidth, color=linecolor, linestyle=linestyle, label=avg_label)
    
    plt.grid(True, linestyle='--', alpha=0.7)
    
    # Add labels
    plt.xlabel(x_name, fontsize=fontsize)
    plt.ylabel(y_name, fontsize=fontsize)
    
    # Handle ticks
    if x_ticks is not None:
        if isinstance(x_ticks, int):
            xmin    = plt.gca().get_xlim()[0]
            xmax    = plt.gca().get_xlim()[1]
            x_ticks = np.linspace(xmin, xmax, x_ticks)
        plt.xticks(ticks=x_ticks)
    if y_ticks is not None:
        if isinstance(y_ticks, int):
            ymin    = plt.gca().get_ylim()[0]
            ymax    = plt.gca().get_ylim()[1]
            y_ticks = np.linspace(ymin, ymax, y_ticks)
        plt.yticks(ticks=y_ticks)
        
    if logx is True:
        plt.xscale('log')
    if logy is True:
        plt.yscale('log')
        
    plt.xticks(fontsize=fontsize*3//4)
    plt.yticks(fontsize=fontsize*3//4)
    
    if Z_st is not None:
        y_min, y_max = plt.gca().get_ylim()
        plt.vlines(Z_st, y_min, y_max, color='k', linestyle='-.', linewidth=2, label=r'$Z_{st}$')
        plt.ylim([y_min, y_max])
        plt.legend()
    
    # Adjust borders tickness
    for spine in plt.gca().spines.values():
        spine.set_linewidth(ParityPlot.border_width )

    if remove_x:
        plt.xlabel('')
        plt.xticks([])
    if remove_y:
        plt.xlabel('')
        plt.xticks([])
    
    figure = plt.gcf()
    
    if save:
        if title is None:
            title=f'Contour_plot'
        else:
            title = 'Conditional_mean_' + title
        plt.savefig(title, dpi=dpi, bbox_inches = "tight")
    
    if show:
        plt.show()
        
    # Close the figure to prevent it from displaying when returned
    plt.close(figure)
    
    return figure

def plot_multifield(X, Y, 
                    fields, 
                    titles=None, 
                    n_cols=None, 
                    figsize=None, 
                    cmap='viridis',
                    vmin=None,
                    vmax=None,
                    side_text=None):
    """
    Plot multiple 2D fields as subplots with shared colorbar and scaling.
    
    Parameters:
    -----------
    X : array-like
        2D mesh for X coordinates
    Y : array-like  
        2D mesh for Y coordinates
    fields : array-like
        3D array of shape (n_fields, ny, nx) or list of 2D arrays
        Each field represents the same variable at different conditions
    titles : list of str, optional
        Titles for each subplot. If None, uses "Field 1", "Field 2", etc.
    n_cols : int, optional
        Number of columns in the subplot grid. If None, automatically determined
    figsize : tuple, optional
        Figure size (width, height). If None, automatically computed based on data aspect ratio
    cmap : str, optional
        Colormap name (default: 'viridis')
    vmin : float, optional
        Minimum value for colorbar. If None, uses global minimum of fields
    vmax : float, optional
        Maximum value for colorbar. If None, uses global maximum of fields
    side_text : str, optional
        Text to display on the right side of the figure
    
    Returns:
    --------
    fig : matplotlib.figure.Figure
        The figure object
    axes : numpy.ndarray
        Array of subplot axes
    """
    
    # Convert fields to numpy array if it's a list
    fields = np.array(fields)
    
    # Get dimensions
    if fields.ndim == 3:
        n_fields, ny, nx = fields.shape
    else:
        raise ValueError("fields must be 3D array (n_fields, ny, nx) or list of 2D arrays")
    
    # Calculate grid layout
    if n_cols is None:
        # Try to make it as square as possible
        n_cols = int(np.ceil(np.sqrt(n_fields)))
    n_rows = int(np.ceil(n_fields / n_cols))
    
    # Calculate data aspect ratio and extent
    x_extent = np.max(X) - np.min(X)
    y_extent = np.max(Y) - np.min(Y)
    data_aspect_ratio = y_extent / x_extent
    
    # Set figure size if not provided
    if figsize is None:
        # Base subplot size considering data aspect ratio
        base_width = 1.5  # Base width for each subplot
        subplot_height = base_width * data_aspect_ratio
        
        # Calculate total figure size
        total_width = base_width * (n_cols*1.1 + 1)  # Extra space for colorbar
        total_height = subplot_height * (n_rows + 0.2)  # Extra space for titles
        
        figsize = (total_width, total_height)
    
    # Create figure and subplots
    fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)
    
    # Handle single subplot case
    if n_fields == 1:
        axes = np.array([axes])
    elif axes.ndim == 1:
        axes = axes.reshape(n_rows, n_cols)
    
    # Calculate global vmin and vmax for consistent scaling
    if vmin is None:
        vmin = np.min(fields)
    if vmax is None:
        vmax = np.max(fields)
    
    # Create default titles if not provided
    if titles is None:
        titles = [f"Field {i+1}" for i in range(n_fields)]
    
    # Plot each field
    images = []
    for i in range(n_fields):
        row = i // n_cols
        col = i % n_cols
        ax = axes[row, col] if n_rows > 1 else axes[col]
        
        # Create the plot
        im = ax.pcolormesh(X, Y, fields[i], vmin=vmin, vmax=vmax, cmap=cmap, shading='auto')
        images.append(im)
        
        # Set equal aspect ratio
        ax.set_aspect('equal', adjustable='box')
        
        # Set title
        ax.set_title(titles[i])
        
        # Handle axis labels based on position
        # Y-axis labels only for leftmost column
        if col == 0:
            ax.set_ylabel('Y [mm]')
        else:
            ax.set_yticklabels([])
        
        # X-axis labels only for bottom row
        if row == n_rows - 1:
            ax.set_xlabel('X [mm]')
        else:
            ax.set_xticklabels([])
    
    # Hide unused subplots
    for i in range(n_fields, n_rows * n_cols):
        row = i // n_cols
        col = i % n_cols
        ax = axes[row, col] if n_rows > 1 else axes[col]
        ax.set_visible(False)
    
    # Add single colorbar on the right
    plt.tight_layout()
    
    # Adjust layout to make room for colorbar
    plt.subplots_adjust(right=0.9)
    
    # Create colorbar axes
    cbar_ax = fig.add_axes([0.94, 0.15, 0.02, 0.7])
    cbar = fig.colorbar(images[0], cax=cbar_ax)
    
    # Add side text if provided (after everything is laid out)
    if side_text is not None:
        # Add text outside the figure area - matplotlib will expand as needed
        fig.text(1.07, 0.5, side_text, 
                rotation=0, verticalalignment='center', 
                fontsize=22, ha='left', transform=fig.transFigure)
    
    return fig, axes

def bins(x, y, n=40, log=False):
    """
    Bins the data in x and y into n bins and computes the average, max, and min of y for each bin.
    
    Args:
    -----
        x (array-like): 
            Independent variable.
        y (array-like): 
            Dependent variable.
        n (int, optional): 
            Number of bins. Default is 40.
        log (bool, optional): 
            If True, use logarithmic bins. Default is False.
    
    Returns:
    --------
        tuple
            - x_b (np.ndarray): Midpoints of the bins in x.
            - y_b (np.ndarray): Average values of y in each bin.
            - y_max (np.ndarray): Maximum values of y in each bin.
            - y_min (np.ndarray): Minimum values of y in each bin.
    """
    x = np.copy(np.asarray(x))
    y = np.copy(np.asarray(y))
    
    x, y = reorder_arrays(x, y)
    
    # Create bins
    if log:
        bins = np.logspace(np.min(x), np.max(x), n + 1)
    else:
        bins = np.linspace(np.min(x), np.max(x), n + 1)
        
    x_b   = []
    y_b   = []
    y_max = []
    y_min = []
    y_var = []
    for i in range(n):
        # check if we have points in the interval
        if np.any((x >= bins[i]) & (x <= bins[i+1])):
            y_bin_i = y[(x >= bins[i]) & (x <= bins[i+1])]
            x_b.append((bins[i+1]+bins[i])/2)
            y_b.append   (np.average(y_bin_i))
            y_max.append (np.max    (y_bin_i))
            y_min.append (np.min    (y_bin_i))
            y_var.append (np.var    (y_bin_i))
    
    x_b   = np.array(x_b)
    y_b   = np.array(y_b)
    y_max = np.array(y_max)
    y_min = np.array(y_min)
        
    return x_b, y_b, y_max, y_min, y_var

def generate_colors(num_colors, cmap='plasma'):
    """
    Generates a list of distinct colors.

    Args:
        num_colors (int): The number of colors to generate.

    Returns:
        list: A list of colors in hexadecimal format.
    """
    # Generate a colormap
    cmap = plt.get_cmap(cmap)
    
    # Generate colors using the colormap
    colors = [cmap(i / num_colors) for i in range(num_colors)]
    
    # Convert RGBA to hex
    hex_colors = [matplotlib.colors.to_hex(color) for color in colors]
    
    return hex_colors

def generate_markers(n):
    all_markers = list(['o', 's', 'D', '^', '>', '<', 'd', 'p', 'H', '*'])
    return all_markers[:n]

def generate_lines(n):
    linestyle_tuple = [
     ('solid',                 '-'),
     ('dashed',                '--'),
     ('dashdotted',            '-.'),
     ('dotted',                (0, (1, 1))),
     ('long dash with offset', (5, (10, 3))),
     ('loosely dashdotdotted', (0, (3, 10, 1, 10, 1, 10))),
     
     ('loosely dotted',        (0, (1, 10))),
     ('dotted',                (0, (1, 1))),
     ('densely dotted',        (0, (1, 1))),
     ('long dash with offset', (5, (10, 3))),
     ('loosely dashed',        (0, (5, 10))),
     ('dashed',                (0, (5, 5))),
     ('densely dashed',        (0, (5, 1))),

     ('loosely dashdotted',    (0, (3, 10, 1, 10))),
     ('dashdotted',            (0, (3, 5, 1, 5))),
     ('densely dashdotted',    (0, (3, 1, 1, 1))),

     ('dashdotdotted',         (0, (3, 5, 1, 5, 1, 5))),
     ('loosely dashdotdotted', (0, (3, 10, 1, 10, 1, 10))),
     ('densely dashdotdotted', (0, (3, 1, 1, 1, 1, 1)))]
    
    lines = [linestyle_tuple[i][1] for i in range(n)]
    return lines


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" A collection of training data building utilities  
"""

__authors__ = "Lorenzo Piu"
__copyright__ = "Copyright (c) 2024-2026, Lorenzo Piu, Heinz Pitsch, and Alessandro Parente"
__credits__ = ["Aero-Thermo-Mechanics laboratories - Universite Libre de Bruxelles, Brussels, Belgium"
               "Institut für Technische Verbrennung (ITV) - RWTH Aachen University, Aachen, Germany"]
__license__ = "GPL 3.0"
__version__ = "1.3.0"
__maintainer__ = ["Lorenzo Piu"]
__email__ = ["lorenzo.piu@ulb.be"]
__status__ = "Production"


import json
import os
from .DNS import Field
import numpy as np
  
class TrainingBuilder(dict):
    """
    Description:
    ------------
    
    A dictionary-like class that builds and manages a collection of VectorScaler instances for training purposes.

    This class includes methods to add, configure, and save/load multiple VectorScaler objects, 
    and allows batch scaling of data based on a specified Field input.
    
    Attributes:
    -----------
        - state_dict (dict): A dictionary containing the state of all scalers in the TrainingBuilder.
        
    Methods:
    --------
        - __init__(*args, **kwargs):
            Initialize the TrainingBuilder with optional dictionary arguments.

        - get_subset(keys):
            Return a subset of the TrainingBuilder based on the specified keys.

        - __setitem__(key, value):
            Set an item in the TrainingBuilder with key and value type validation.

        - add(variable, *args, **kwargs):
            Add a new VectorScaler to the TrainingBuilder.

        - build_x(field):
            Construct a feature matrix by transforming data from a Field instance.

        - fit(field):
            Fit each VectorScaler in the TrainingBuilder to the corresponding data in the Field.

        - load(file_path):
            Load the state of each scaler from a JSON file and reinitialize the TrainingBuilder.

        - save(file_path):
            Save the current state of the TrainingBuilder to a JSON file.

        - state_dict:
            Return the current state of the TrainingBuilder, including all scaler states.

        - __str__():
            Return a string representation of the TrainingBuilder.
    """
    
    def __init__(self, *args, **kwargs):
        """
        Initialize the TrainingBuilder with optional dictionary arguments.

        Parameters:
        -----------
            - *args: 
                Variable length argument list. Must be coherent with the input arguments used
                in the initialization of a dictionary.
                
            - **kwargs: 
                Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)

    def get_subset(self, keys):
        """
        Return a subset of the TrainingBuilder based on the specified keys.

        Parameters:
        -----------
            - keys (list): 
                The keys to include in the subset.

        Returns:
        --------
            - TrainingBuilder: 
                A new TrainingBuilder containing only the specified keys.
        """
        return TrainingBuilder({k: self[k] for k in keys if k in self})

    def __setitem__(self, key, value):
        """
        Set an item in the TrainingBuilder with key validation.

        Parameters:
        -----------
            - key (str): 
                The key for the item.
                
            - value (VectorScaler): 
                The VectorScaler object to add.

        Raises:
        -------
            - KeyError: If the key is not a string.
            - ValueError: If the value is not an instance of VectorScaler.
        """
        if not isinstance(key, str):
            raise KeyError("Keys must be strings.")
        if not isinstance(value, VectorScaler):  # Example type restriction
            raise ValueError("Values must be int or str.")
        super().__setitem__(key, value)

    def add(self, variable, *args, **kwargs):
        """
        Add a new VectorScaler to the TrainingBuilder.

        Parameters:
        -----------
            - variable (str): 
                The key to associate with the VectorScaler.
                
            - *args: 
                Arguments to pass to the VectorScaler initializer.
            
            - **kwargs: 
                Keyword arguments to pass to the VectorScaler initializer.

        Raises:
        -------
            - KeyError: 
                If the specified variable already exists in the TrainingBuilder.
        """
        if variable in self:
            raise KeyError(f"Key '{variable}' already exists.")
        
        value = VectorScaler(*args,**kwargs)
        self[variable] = value  # Calls __setitem__
        
    def build_x(self, field):
        """
        Construct a feature matrix by transforming data from a Field instance.

        Parameters:
        -----------
            - field (Field): The input field containing data to scale.

        Returns:
        --------
            - np.ndarray: A 2D array containing the scaled data for all variables in the TrainingBuilder.

        Raises:
        -------
            - KeyError: If the field is not an instance of Field.
        """
        def append_column(array_2d, column):
            column = np.array(column).reshape(-1, 1)  # Reshape to a 2D column
            return np.hstack((array_2d, column)) if array_2d.size else column  # Append column or return the new column

        # Check type
        if not isinstance(field, Field):
            raise KeyError("Keys must be strings.")
            
        # Start looping through the variables
        X = np.empty((0,0))
        for variable_name, scaler in self.items():
            scalar = getattr(field, variable_name).value
            scaler.fit(scalar)
            x = scaler.transform(scalar)
            X = append_column(X, x)
            
        return X
    
    # def fit(self, field):
    #     """
    #     Old function, only able to handle one field at a time.
    #     Delete after testing the new utility.
    #     Fit each VectorScaler in the TrainingBuilder to the corresponding data in the Field.

    #     Parameters:
    #     -----------
    #         - field (Field): The field with data to fit each scaler.
    #     """
    #     for variable_name in self.keys():
    #         self[variable_name].fit(getattr(field, variable_name).value)
    
    def fit(self, field):
        """
        Fit each VectorScaler in the TrainingBuilder to the corresponding data in the Field or list of Field objects.

        Parameters:
        -----------
            - field (Field or list of Field): The field(s) with data to fit each scaler.
        """
        # Handle single field case
        if isinstance(field, Field):
            fields = [field]
        # Handle list of fields case
        elif isinstance(field, list) and all(isinstance(f,Field) for f in field):
            fields = field
        else:
            raise TypeError("Input must be a Field object or a list of Field objects")
        
        # Collect data from all fields for each variable
        for variable_name in self.keys():
            # Gather data from all fields for this variable
            all_data = []
            for f in fields:
                variable_data = getattr(f, variable_name).value
                all_data.append(variable_data.flatten())  # Flatten to 1D array
            
            # Concatenate all data for this variable
            combined_data = np.concatenate(all_data)
            
            # Fit the scaler to the combined data
            self[variable_name].fit(combined_data)
            
    def load(self, file_path):
        """
        Load the state of a previously saved object from a JSON file and reinitialize the TrainingBuilder.

        Parameters:
        -----------
            - file_path (str): The path to the JSON file to load.

        Raises:
        -------
            - ValueError: If the file is not a valid JSON file.
        """
        # Check that the file is a json
        if not is_json_file(file_path):
            raise ValueError(f"The file {file_path} is not a valid json file.")
            
        # Read the dictionary
        with open(file_path, 'r') as file:
            state_dict = json.load(file)
        
        # reinitialize object
        self.clear()
        for variable, state_dict_scaler in state_dict.items():
            scaler = VectorScaler()
            print(state_dict_scaler)
            scaler.load(state_dict_scaler)
            self[variable] = scaler
            
    def save(self, file_path):
        """
        Save the current state of the TrainingBuilder to a JSON file.

        Parameters:
        -----------
            - file_path (str): The path to save the JSON file.
        """
        with open(file_path, 'w') as file:
            json.dump(self.state_dict, file, indent=4)
        
    @property
    def state_dict(self):
        """
        Return the current state of the TrainingBuilder, including all scaler states.

        Returns:
        --------
            - dict: A dictionary representing the state of all scalers in the TrainingBuilder.
        """
        state_dict = dict()
        for key in self.keys():
            state_dict[key] = self[key].state_dict
        return state_dict

    def __str__(self):
        """
        Return a string representation of the TrainingBuilder.

        Returns:
        --------
            - str: The string representation.
        """
        return f"TrainingBuilder({super().__str__()})"
    

class VectorScaler():
    """
    Description:
    ------------
    A class for scaling and transforming vector data. This class supports various scaling
    modes, such as min-max scaling, standard scaling, and mean scaling, with additional options
    for modulus transformation, logarithmic scaling, and value clipping.
    
    The operations are performed in the following order:
        1. Modulus transformation
        2. Clipping
        3. Logarithmic transformation (with an automatic clipping at 1e-20 to avoid negative or zero values)
        4. Scaling based on the specified mode
    
    Attributes:
    -----------
        - mode (str): 
            The scaling mode. Options are 'minmax', 'standard', 'mean', or None.
            
        - modulus (bool): 
            Whether to apply the modulus (absolute value) operation.
        
        - log (bool): 
            Whether to apply logarithmic transformation.
        
        - vmin (float, optional): 
            Minimum value for clipping.
        
        - vmax (float, optional): 
            Maximum value for clipping.
        
        - copy (bool): 
            Whether to create a copy of the input array.
        
        - max, min, mean, std, ptp (float): 
            calculated scaling parameters, depending on mode.
        
    Methods:
    --------
        - __init__(mode='minmax', modulus=False, log=False, vmin=None, vmax=None, copy=True):
            Initialize the VectorScaler with scaling mode, transformation options, and optional clipping bounds.

        - _preprocess_input(x):
            Preprocess the input array by applying modulus, clipping, and logarithmic transformations.

        - _reset():
            Reset the scaler attributes (max, min, mean, std, ptp) to None. Used before recalculating parameters.

        - fit(x):
            Fit the VectorScaler to the data by calculating the necessary statistics based on the specified mode.

        - transform(x):
            Transform the input data based on the fitted scaling parameters and mode.

        - load(state_dict):
            Load saved state values into the VectorScaler.

        - state_dict:
            Return the current state of the VectorScaler, including mode and calculated parameters.
    """
    _lower_limit = 1e-30
    
    _args = ['mode', 'modulus', 'log', 'vmin', 'vmax', 'copy']
    _scalers = ['max', 'min', 'mean', 'std', 'ptp']
    
    def __init__(self, mode='minmax', modulus=False, log=False, vmin=None, vmax=None, copy=True):
        """
        Initialize the VectorScaler with scaling mode, transformation options, and optional clipping bounds.

        Parameters:
        -----------
            - mode (str): 
                The scaling mode. Options are 'minmax', 'standard', 'mean', or None.
                
            - modulus (bool): 
                Whether to apply modulus transformation.
                
            - log (bool): 
                Whether to apply logarithmic transformation.
                
            - vmin (float, optional): 
                Minimum value for clipping.
                
            - vmax (float, optional): 
                Maximum value for clipping.
                
            - copy (bool): 
                Whether to create a copy of the input array.
        """
        self.mode    = mode
        self.modulus = modulus
        self.log     = log
        self.vmin    = vmin
        self.vmax    = vmax
        self.copy    = copy
        
        self.max = None
        self.min = None
        self.mean = None
        self.std = None
        self.ptp = None
        
    def _preprocess_input(self, x):
        """
        Preprocess the input array by applying modulus, clipping, and logarithmic transformations.

        Parameters:
        -----------
            - x (array-like): 
                The input data to preprocess.

        Returns:
        --------
            - np.ndarray: 
                The preprocessed input as a 2D column vector.
        """
        # Copy if needed
        if self.copy:
            x = x.copy()
            
        # Convert to float array and reshape as a column vector
        x = np.array(x).flatten().astype(float)
        x = np.reshape(x, [len(x), 1])

        # Apply modulus if necessary
        if self.modulus:
            x = np.abs(x)

        # Clip values based on vmin and vmax
        if self.vmin is not None:
            x[x < self.vmin] = self.vmin
        if self.vmax is not None:
            x[x > self.vmax] = self.vmax

        # Apply logarithmic transformation if enabled
        if self.log:
            if any(x < VectorScaler._lower_limit):
                x[x < VectorScaler._lower_limit] = VectorScaler._lower_limit
            x = np.log10(x)
            
        return x
    
    def _reset(self):
        """
        Reset the scaler attributes (max, min, mean, std, ptp) to None. Used before recalculating parameters.
        """
        self.max = None
        self.min = None
        self.mean = None
        self.std = None
        self.ptp = None
    
    def fit(self, x):
        """
        Fit the VectorScaler to the data by calculating the necessary statistics based on the specified mode.

        Parameters:
        -----------
            - x (array-like): 
                The data to fit the scaler to.

        Raises:
        -------
            - ValueError: 
                If the mode is unrecognized.
        """
        self._reset()
        
        x = self._preprocess_input(x)
        
        if self.mode is None:
            pass
    
        elif self.mode.lower()=='minmax':
            self.max = np.max(x)
            self.min = np.min(x)
            
        elif self.mode.lower() == 'standard':
            self.mean = np.mean(x)
            self.std = np.std(x)
            
        elif self.mode.lower() == 'mean':
            self.mean = np.mean(x)
            self.ptp = np.ptp(x)
        
        else:
            raise ValueError(f"Unknown scaling mode {self.mode}")
            
    
    def transform(self, x):
        """
        Transform the input data based on the fitted scaling parameters and mode.

        Parameters:
        -----------
            - x (array-like): 
                The data to transform.

        Returns:
        --------
            - np.ndarray: 
                The scaled and transformed data.
        
        Raises:
        -------
            - ValueError: 
                If the mode is unrecognized.
        """
        x = self._preprocess_input(x)
        
        if self.mode is None:
            pass

        elif self.mode.lower()=='minmax':
            x = (x-self.min)/(self.max-self.min)
            
        elif self.mode.lower() == 'standard':
            x = (x - self.mean) / self.std
            
        elif self.mode.lower() == 'mean':
            # Scale between -1 and 1 using mean and range
            x = (x - self.mean) / (self.ptp / 2)
        
        else:
            raise ValueError(f"Unknown scaling mode {self.mode}")

        return x
    
    def load(self, state_dict):
        """
        Load saved state values into the VectorScaler.

        Parameters:
        -----------
            - state_dict (dict): 
                Dictionary containing scaler attributes to load.
        """
        for arg in state_dict.keys():
            setattr(self, arg, state_dict[arg])
    
    @property
    def state_dict(self):
        """
        Return the current state of the VectorScaler, including mode and calculated parameters.

        Returns:
        --------
            - dict: 
                A dictionary representation of the scaler's current state.
        """
        state_dict = dict()
        for arg in VectorScaler._args:
            state_dict[arg] = getattr(self, arg)
        for arg in VectorScaler._scalers:
            if getattr(self, arg) is not None:
                state_dict[arg] = getattr(self, arg)
        return state_dict
    

def is_json_file(file_path):
    """
    Check if the given file_path is a JSON file and if it contains valid JSON data.

    Parameters:
    -----------
        - file_path (str): The path to the file.

    Returns:
    --------
        - bool: True if the file is a valid JSON file, False otherwise.
    """
    # Check for .json file extension
    if not file_path.endswith('.json'):
        return False

    # Check if the file exists
    if not os.path.isfile(file_path):
        return False

    # Try to open and load the JSON content
    try:
        with open(file_path, 'r') as file:
            json.load(file)  # This will raise an error if the file is not valid JSON
    except (json.JSONDecodeError, OSError):
        return False

    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 24 14:50:44 2024

@author: lorenzo piu
"""

import os
import aPriori as ap
from aPriori.DNS import Field

# Uncomment the following line if you did not download the dataset yet
# ap.download(dataset='h2_lifted')

directory = os.path.join('.', 'Lifted_H2_subdomain') # change this with your path to the data folder

T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path \'{T_path}\' is correct...")
if not os.path.exists(T_path):
    print(f"The path '{T_path}' does not exist in your system. Downloading the dataset from Github...")
    ap.download(dataset='h2_lifted')

# Initialize the DNS field
DNS_field = Field(directory)

# Compute the reaction rates
DNS_field.compute_reaction_rates()

# # Compute the strain rate module on the DNS data without saving the tensor components
DNS_field.compute_strain_rate(save_tensor=False)

# FILTERING
filter_size = 16

# filter DNS field and initialize filtered field
filtered_field = Field(DNS_field.filter_favre(filter_size))

# compute the reaction rates on the filtered field
filtered_field.compute_reaction_rates()

# compute the strain rate on the filtered field
filtered_field.compute_strain_rate(save_tensor=True)

# compute the residual dissipation rate with Smagorinsky model
filtered_field.compute_residual_dissipation_rate(mode='Smag')

# compute residual kinetic energy
filtered_field.compute_residual_kinetic_energy()

# compute chemical timescale with SFR, FFR and Chomiak model 
filtered_field.compute_chemical_timescale(mode='SFR')
filtered_field.fuel = 'H2'
filtered_field.ox = 'O2'
filtered_field.compute_chemical_timescale(mode='Ch')
filtered_field.compute_mixing_timescale(mode='Kolmo')

###############################################################################
#                        Machine Learning Closure
###############################################################################
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split

# DATA PROCESSING
T      = filtered_field.T.reshape_column() # extract the valeue of Temperature from the filtered field and rehape it as a column vector
S      = filtered_field.S_LES.reshape_column()
Tau_c  = filtered_field.Tau_c_SFR.reshape_column()

HRR_LFR = filtered_field.HRR_LFR.reshape_column()
HRR_DNS = filtered_field.HRR_DNS.reshape_column()

# Data scaling
T      = (T-np.min(T))/(np.max(T)-np.min(T))
S      = np.log10(S)
S      = (S-np.min(S))/(np.max(S) - np.min(S))
Tau_c  = np.log10(Tau_c)
Tau_c  = (Tau_c-np.min(Tau_c)) / (np.max(Tau_c)-np.min(Tau_c))

# Build the training data matrix
X = np.hstack([T, S, Tau_c])

# Divide between train and test data
X_train, X_test, HRR_LFR_train, HRR_LFR_test, HRR_DNS_train, HRR_DNS_test = train_test_split(
    X, HRR_LFR, HRR_DNS, test_size=0.9, random_state=42)

X_train = torch.tensor(X_train, dtype=torch.float32)
X_test  = torch.tensor(X_test,  dtype=torch.float32)
HRR_LFR_train = torch.tensor(HRR_LFR_train, dtype=torch.float32)
HRR_LFR_test  = torch.tensor(HRR_LFR_test,  dtype=torch.float32)
HRR_DNS_train = torch.tensor(HRR_DNS_train, dtype=torch.float32)
HRR_DNS_test  = torch.tensor(HRR_DNS_test,  dtype=torch.float32)


# NN class definition
class NeuralNetwork(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, num_layers):
        super(NeuralNetwork, self).__init__()
        self.layers = nn.ModuleList() # initialize the layers list as an empty list using nn.ModuleList()
        self.layers.append(nn.Linear(input_size, hidden_size)) # Add the first input layer. The layer takes as input <input_size> neurons and gets as output <hidden_size> neurons
        for _ in range(num_layers - 1):
            self.layers.append(nn.Linear(hidden_size, hidden_size)) # Add hidden layers
        self.layers.append(nn.Linear(hidden_size, output_size)) # add output layer

    def forward(self, x):    # Function to perform forward propagation
        for layer in self.layers[:-1]:
            x = torch.relu(layer(x))
        x = self.layers[-1](x)
        return x
    
# NN architecture
input_size = 3
num_layers = 6
hidden_size = 64
output_size = 1
model = NeuralNetwork(input_size, hidden_size, output_size, num_layers)

# Define loss function and optimizer
criterion = nn.MSELoss()
optimizer = optim.Adam(model.parameters()) # here we are using the Adam optimizer, to optimize model.parameters, but what is there inside this attribute?

# transfer on GPU
if torch.backends.mps.is_available():
    device = "mps"
elif torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"
    
model = model.to(device)
X_train = X_train.to(device)
X_test = X_test.to(device)
HRR_LFR_train = HRR_LFR_train.to(device)
HRR_DNS_train = HRR_DNS_train.to(device)
HRR_LFR_test = HRR_LFR_test.to(device)
HRR_DNS_test = HRR_DNS_test.to(device)

# Move the optimizer's state to the same device
for state in optimizer.state.values():
    for k, v in state.items():
        if isinstance(v, torch.Tensor):
            state[k] = v.to(device)

# Lists to store training loss and testing loss
train_loss_list = []
test_loss_list = []

# Training loop
num_epochs = 1000
for epoch in range(num_epochs):
    # Forward pass
    output = model(X_train)

    # compute loss on training data
    loss = criterion(output*HRR_LFR_train, HRR_DNS_train)

    # Compute loss on testing data. NOTE: we aren't gonna use the test loss for optimization!!!
    with torch.no_grad():
        output_test = model(X_test)
        loss_test = criterion(output_test*HRR_LFR_test, HRR_DNS_test)

    # Backprop and optimize
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()

    train_loss_list.append(loss.item()) # Save the losses
    test_loss_list.append(loss_test.item())

    if (epoch + 1) % 10 == 0:
        print(f'Epoch [{epoch+1}/{num_epochs}], Loss: {loss.item():.4f}, Test loss: {loss_test.item():.4f}')

# Plot training and testing loss
plt.plot(train_loss_list, label='Training Loss')
plt.plot(test_loss_list, label='Testing Loss')
plt.xlabel('Epoch')
plt.ylabel('Loss')
plt.title('Training and Testing Loss')
plt.yscale('log')
plt.legend()
plt.show()

# PLOTTING
with torch.no_grad():
    gamma = model(torch.tensor(X, dtype=torch.float32).to(device)).cpu().numpy()

# Visualize the results
f = ap.parity_plot(HRR_DNS, HRR_LFR, density=True, 
               x_name=r'$\dot{Q}_{DNS}$',
               y_name=r'$\dot{Q}_{LFR}$',
               cbar_title=r'$\rho_{KDE}/max(\rho_{KDE})$',
               )
f = ap.parity_plot(HRR_DNS, gamma*HRR_LFR,density=True, 
               x_name=r'$\dot{Q}_{DNS}$',
               y_name=r'$\dot{Q}_{ML}$',
               cbar_title=r'$\rho_{KDE}/max(\rho_{KDE})$',
               )

gamma_2D = gamma.reshape(filtered_field.shape)[:,:,filtered_field.shape[2]//2] # extract the z midplane of gamma
HRR_LFR_2D = HRR_LFR.reshape(filtered_field.shape)[:,:,filtered_field.shape[2]//2]# extract the z midplane
HRR_ML_2D = gamma_2D * HRR_LFR_2D
HRR_DNS_2D = HRR_DNS.reshape(filtered_field.shape)[:,:,filtered_field.shape[2]//2]# extract the z midplane

f = ap.contour_plot(filtered_field.mesh.X_midZ*1000,   # Extract x mesh on the z midplane
                    filtered_field.mesh.Y_midZ*1000,   # Extract y mesh on the z midplane
                    np.abs(HRR_LFR_2D-HRR_DNS_2D),
                    vmax=1.5e10,
                    colormap='Reds',
                    x_name='x [mm]',
                    y_name='y [mm]',
                    title=r'$|\dot{Q}_{LFR}-\dot{Q}_{DNS}|$'
                    )

f = ap.contour_plot(filtered_field.mesh.X_midZ*1000,   # Extract x mesh on the z midplane
                    filtered_field.mesh.Y_midZ*1000,   # Extract y mesh on the z midplane
                    np.abs(HRR_ML_2D-HRR_DNS_2D),
                    vmax=1.5e10,
                    colormap='Reds',
                    x_name='x [mm]',
                    y_name='y [mm]',
                    title=r'$|\dot{Q}_{LFR}-\dot{Q}_{DNS}|$'
                    )

# Visualize the NN output
f = ap.contour_plot(filtered_field.mesh.X_midZ*1000,   # Extract x mesh on the z midplane
                    filtered_field.mesh.Y_midZ*1000,   # Extract y mesh on the z midplane
                    gamma_2D,
                    colormap='viridis',
                    x_name='x [mm]',
                    y_name='y [mm]',
                    title=r'$\gamma_{NN}$'
                    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Library initialization and relative imports
"""

__author__ = "Lorenzo Piu"
__copyright__ = "Copyright (c) 2024-2026, Lorenzo Piu"
__credits__ = ["Universite Libre de Bruxelles, Aero-Thermo-Mechanics Laboratory, Brussels, Belgium"]
__license__ = "GPL 3.0"
__version__ = "1.3.0"
__maintainer__ = ["Lorenzo Piu"]
__email__ = ["lorenzo.piu@ulb.be"]
__status__ = "Production"

# Import useful functions and classes from the DNS module
from .DNS import (
                 add_variable,
                 delete_file, 
                 download, 
                 process_file, 
                 filter_gauss, 
                 filter_box, 
                 filter_3D, 
                 save_file, 
                 gradient_x, 
                 gradient_y, 
                 gradient_z,
                 Field,
                 Scalar,
                 Mesh
                 )

# Import plot utilities
from .plot_utilities import (contour_plot,
                             parity_plot,
                             cond_mean_plot,
                             scatter
                             )

# # Default modules to import
# import inspect
# import importlib

# # Function to dynamically import all functions and classes from a given module
# def import_module_contents(module_name):
#     module = importlib.import_module(module_name, package=__name__)
#     for name, member in inspect.getmembers(module):
#         if inspect.isfunction(member) or inspect.isclass(member):
#             globals()[name] = member
#             __all__.append(name)
            
# def import_module_classes(module_name):
#     module = importlib.import_module(module_name, package=__name__)
#     for name, member in inspect.getmembers(module):
#         if inspect.isclass(member):
#             globals()[name] = member
#             __all__.append(name)

# # Initialize __all__ list
# __all__ = []

# # Import all the classes from the module DNS
# import_module_classes('.DNS')

# # Import contents from the module plot_utilities
# import_module_contents('.plot_utilities')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""

folder_structure = {
                     "data_path" : "data",
                     "grid_path" : "grid",
                     "chem_path" : "chem_thermo_tran"
                 }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""List of attributes available for data processing
"""

__authors__ = "Lorenzo Piu"
__copyright__ = "Copyright (c) 2024-2026, Lorenzo Piu, Heinz Pitsch, and Alessandro Parente"
__credits__ = ["Aero-Thermo-Mechanics laboratories - Universite Libre de Bruxelles, Brussels, Belgium"
               "Institut für Technische Verbrennung (ITV) - RWTH Aachen University, Aachen, Germany"]
__license__ = "GPL 3.0"
__version__ = "1.3.0"
__maintainer__ = ["Lorenzo Piu"]
__email__ = ["lorenzo.piu@ulb.be"]
__status__ = "Production"

variables_list = {
           # Attribute Name   :           File Name             :  Species  :         Models          :      Tensor      :     Reactions      :     Variable Description
             "dU{}_dX{}_{}"   : ["dU{}_dX{}_{}_s-1_{}.dat"      ,   False   ,    ['DNS', 'LES']       ,      True        ,       False        , "Derivative with respect to j of the ith velocity component"],
             "S_{}"           : ["S_{}_s-1_{}.dat"              ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Module of the Strain Rate tensor"],
             "S{}{}_{}"       : ["S{}{}_{}_s-1_{}.dat"          ,   False   ,    ['DNS', 'LES']       ,   'Symmetric'    ,       False        , "Module of the Strain Rate tensor"],
             "P"              : ["P_Pa_{}.dat"                  ,   False   ,       None              ,      False       ,       False        , "Pressure"],
             "RHO"            : ["RHO_kgm-3_{}.dat"             ,   False   ,       None              ,      False       ,       False        , "Density"],
             "T"              : ["T_K_{}.dat"                   ,   False   ,       None              ,      False       ,       False        , "Temperature"],
             "U_{}"           : ["U_{}_ms-1_{}.dat"             ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Module of the velocity vector"],
             "U_X"            : ["UX_ms-1_{}.dat"               ,   False   ,       None              ,      False       ,       False        , "X component of the velocity"],
             "U_Y"            : ["UY_ms-1_{}.dat"               ,   False   ,       None              ,      False       ,       False        , "Y component of the velocity"],
             "U_Z"            : ["UZ_ms-1_{}.dat"               ,   False   ,       None              ,      False       ,       False        , "Z component of the velocity"],
             "Y{}"            : ["Y{}_{}.dat"                   ,   True    ,       None              ,      False       ,       False        , "Species' mass fraction; expected one scalar field for every specie in the chemical mechanism"],
             "R{}_LFR"        : ["R{}_LFR_kgm-3s-1_{}.dat"      ,   True    ,       None              ,      False       ,       False        , "Species' Reaction Rates, computed as the Arrhenius rates on the filtered values of T, Y; expected one scalar field for every specie in the chemical mechanism"],
             "R{}_DNS"        : ["R{}_DNS_kgm-3s-1_{}.dat"      ,   True    ,       None              ,      False       ,       False        , "Species' Reaction Rates, computed as the Arrhenius rates on the DNS values of T, Y, and eventually filtered; expected one scalar field for every specie in the chemical mechanism"],
             "R{}_Batch"      : ["R{}_Batch_kgm-3s-1_{}.dat"    ,   True    ,       None              ,      False       ,       False        , "Species' Reaction Rates, computed as the PaSR rates on the DNS values of T, Y, and eventually filtered; expected one scalar field for every specie in the chemical mechanism"],
             "Tau_c_{}"       : ["Tau_c_{}_s_{}.dat"            ,   False   , ['Ch', 'SFR', 'FFR']    ,      False       ,       False        , "Chemical timescale. Models available: Chomiak ('Ch'), Slowest Formation Rate ('SFR'), Fastest Formation Rate ('FFR')"],
             "Tau_m_{}"       : ["Tau_m_{}_s_{}.dat"            ,   False   , ['Kolmo', 'Int', 'Sub'] ,      False       ,       False        , "Mixing timescale. Models available: Kolmogorov ('kolmo'), Integral ('int'), Subgrid velocity stretch ('sub')"],
             "TAU_r_{}{}_{}"  : ["TAU_r_{}{}_{}_m2s-2_{}.dat"   ,   False   , ['DNS', 'Smag']         ,   'Symmetric'    ,       False        , "Anisotropic part of the Residual stress tensor (filt(Ui_DNS*Uj_DNS)-(Ui_LES*Uj_LES) - 2/3*K_res. Models available: DNS ('DNS'), Smagorinsky ('Smag')"],
             "TAU_r_{}"       : ["TAU_r_{}_m2s-2_{}.dat"        ,   False   , ['DNS', 'Smag']         ,      False       ,       False        , "MODULE of the Anisotropic part of the Residual stress tensor. Models available: DNS ('DNS'), Smagorinsky ('Smag')"],
             "Epsilon"        : ["Epsilon_m2s-3_{}.dat"         ,   False   ,       None              ,      False       ,       False        , "Turbulence Dissipation Rate"],
             "Epsilon_r_{}"   : ["Epsilon_m2s-3_{}_{}.dat"      ,   False   , ['DNS', 'Smag']         ,      False       ,       False        , "Turbulence Dissipation Rate"],
             "K_{}"           : ["K_{}_m2s-2_{}.dat"            ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Turbulence Kinetic Energy"],
             "K_r_{}"         : ["K_r_{}_m2s-2_{}.dat"          ,   False   ,    ['DNS', 'Yosh']      ,      False       ,       False        , "Residual Turbulence Kinetic Energy"],
             "HRR_{}"         : ["HRR_{}_kgm2s-3_DNS_{}.dat"    ,   False   , ['DNS', 'LFR','Batch']  ,      False       ,       False        , "Heat Release Rate, computed on unfiltered DNS data"],
             "Mu"             : ["Mu_kgm-1s-1_{}.dat"           ,   False   ,       None              ,      False       ,       False        , "Dynamic Viscosity"],
             "Cp"             : ["Cp_W_m-1_K-1_{}.dat"          ,   False   ,       None              ,      False       ,       False        , "Specific heat"],
             "Lambda"         : ["Lambda_W_m-1_K-1_{}.dat"      ,   False   ,       None              ,      False       ,       False        , "Thermal conductivity"],
             "C"              : ["C_{}.dat"                     ,   False   ,       None              ,      False       ,       False        , "Progress variable"],
             "C_grad"         : ["C_grad_m-1_{}.dat"            ,   False   ,       None              ,      False       ,       False        , "Progress variable gradient (module)"],
             "C_grad_X"       : ["C_grad_X_m-1_{}.dat"          ,   False   ,       None              ,      False       ,       False        , "Progress variable gradient (x component)"],
             "C_grad_Y"       : ["C_grad_Y_m-1_{}.dat"          ,   False   ,       None              ,      False       ,       False        , "Progress variable gradient (y component)"],
             "C_grad_Z"       : ["C_grad_Z_m-1_{}.dat"          ,   False   ,       None              ,      False       ,       False        , "Progress variable gradient (z component)"],
             "C_laplacian"    : ["C_laplacian_m-2_{}.dat"       ,   False   ,       None              ,      False       ,       False        , "Progress variable laplacian"],
             "Z"              : ["Z_{}.dat"                     ,   False   ,       None              ,      False       ,       False        , "Mixture fraction"],
             "Z_grad"         : ["Z_grad_m-1_{}.dat"            ,   False   ,       None              ,      False       ,       False        , "Mixture fraction gradient modulus"],
             "Chi_Z"          : ["Chi_Z_ms-1_{}.dat"            ,   False   ,       None              ,      False       ,       False        , "Mixture fraction Dissipation rate"],
             "M"              : ["M_{}.dat"                     ,   False   ,       None              ,      False       ,       False        , "Fraction of resolved kinetic energy"],
             "PHI_T_X_{}"     : ["PHI_T_X_{}_kgKm-2s-1_{}.dat"  ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Temperature convective flux in the X direction"],
             "PHI_T_Y_{}"     : ["PHI_T_Y_{}_kgKm-2s-1_{}.dat"  ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Temperature convective flux in the Y direction"],
             "PHI_T_Z_{}"     : ["PHI_T_Z_{}_kgKm-2s-1_{}.dat"  ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Temperature convective flux in the Z direction"],
             "TAU_T_X"        : ["TAU_T_X_kgKm-2s-1_{}.dat"     ,   False   ,       None              ,      False       ,       False        , "Sub-filter temperature flux in the X direction"  ],
             "TAU_T_Y"        : ["TAU_T_Y_kgKm-2s-1_{}.dat"     ,   False   ,       None              ,      False       ,       False        , "Sub-filter temperature flux in the Y direction"  ],
             "TAU_T_Z"        : ["TAU_T_Z_kgKm-2s-1_{}.dat"     ,   False   ,       None              ,      False       ,       False        , "Sub-filter temperature flux in the Z direction"  ],
             "PHI_C_X_{}"     : ["PHI_C_X_{}_kgm-2s-1_{}.dat"   ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "PV convective flux in the X direction"],
             "PHI_C_Y_{}"     : ["PHI_C_Y_{}_kgm-2s-1_{}.dat"   ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "PV convective flux in the Y direction"],
             "PHI_C_Z_{}"     : ["PHI_C_Z_{}_kgm-2s-1_{}.dat"   ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "PV convective flux in the Z direction"],
             "TAU_C_X"        : ["TAU_C_X_kgm-2s-1_{}.dat"      ,   False   ,       None              ,      False       ,       False        , "Sub-filter PV flux in the X direction"  ],
             "TAU_C_Y"        : ["TAU_C_Y_kgm-2s-1_{}.dat"      ,   False   ,       None              ,      False       ,       False        , "Sub-filter PV flux in the Y direction"  ],
             "TAU_C_Z"        : ["TAU_C_Z_kgm-2s-1_{}.dat"      ,   False   ,       None              ,      False       ,       False        , "Sub-filter PV flux in the Y direction"  ],
             "TSR_{}"         : ["TSR_{}_{}.dat"                ,   False   ,    ['DNS', 'LES']       ,      False       ,       False        , "Tangential Stretching Rate"  ],
             "Y"              : ["Y_{}.dat"                     ,   False   ,       None              ,      False       ,       False        , "Generic transported scalar"  ],
             "API_TSR_{}"     : ["API_TSR_{}_{}.dat"            ,   False   ,       None              ,      False       ,       True         , "Participation indexes to the TSR"],
             "API_{}_{}"      : ["API_{}_{}_{}.dat"             ,   True    ,       None              ,      False       ,       True         , "Participation indexes on species production rates"],
            }

mesh_list  = {
             "X_mesh"         : ["X_m.dat"                    ,   False   ,       None              , ""],
             "Y_mesh"         : ["Y_m.dat"                    ,   False   ,       None              , ""],
             "Z_mesh"         : ["Z_m.dat"                    ,   False   ,       None              , ""]
             }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 20 14:21:37 2024

@author: lorenzo piu
"""

import os
import aPriori as ap
import json

# Uncomment the following line if you did not download the dataset yet
# ap.download(dataset='h2_lifted')

directory = os.path.join('.','Lifted_H2_subdomain') # change this with your path to the data folder

# Check the folder with the data exists in your system
T_path = os.path.join(directory,'data', 'T_K_id000.dat')
print(f"\nChecking the path \'{T_path}\' is correct...")
if not os.path.exists(T_path):
    raise ValueError("The path '{T_path}' does not exist in your system. Check to have the correct path to your data folder in the code")
else:
    print("Folder path OK\n")

DNS_field = ap.Field(directory)

# Visualize the data
# Default plotting of a variable along the x, y, z midplanes
DNS_field.plot_z_midplane('YH2O2')
DNS_field.plot_y_midplane('YH2O2')
DNS_field.plot_x_midplane('YH2O2')

# Advanced settings to plot. We'll consider the z midplane for simplicity
DNS_field.plot_z_midplane('YOH',
                          levels=[3e-3,10e-3,15e-3], # isocontour lines
                          vmin=3e-3,                 # minimum temperature to plot
                          title=r'$Y_{H2O2}$',       # figure title
                          linewidth=1,               # isocontour lines thickness
                          transpose=True,            # inverts x and y axes
                          x_name='y [mm]',           # x axis label
                          y_name='x [mm]',           # y axis label
                          colormap='inferno',        # change colormap
                          )

# Scatter variables as functions of the mixture fraction z
# Compute the mixture fraction with the compute_mixture_fraction method
DNS_field.ox = 'O2'     # Defines the species to consider as oxydizer
DNS_field.fuel = 'H2'   # Defines the species to consider as fuel
Y_ox_2=0.233  # Oxygen mass fraction in the oxydizer stream (air)
Y_f_1=0.65*2/(0.65*2+0.35*28) # Hydrogen mass fraction in the fuel stream
# (the fuel stream is composed by X_H2=0.65 and X_N2=0.35)

DNS_field.compute_mixture_fraction(Y_ox_2=Y_ox_2, Y_f_1=Y_f_1, s=2)

# Scatter plot variables as functions of the mixture fraction Z
DNS_field.scatter_Z('T', # the variable to plot on the y axis
                    c=DNS_field.YOH.value, # set color of the points
                    y_name='T [K]', 
                    cbar_title=r'$Y_{OH}$'
                    )

DNS_field.scatter_Z('YH2O2', # the variable to plot on the y axis
                    c=DNS_field.YOH.value, # set color of the points
                    y_name=r'$Y_{H2O2}$', 
                    cbar_title=r'$Y_{OH}$'
                    )



# tests/test_h2_premixed_pipeline.py
import os
import numpy as np
import shutil
import pytest

import aPriori as ap

@pytest.mark.slow
@pytest.mark.integration
def test_h2_premixed_end_to_end(apriori_test_cache_dir: str):
    """
    Please run the test with the following command:
      pytest --html=pytest_report.html --self-contained-html

    End-to-end smoke/integration test:
      - download dataset if needed
      - load Field
      - compute reaction rates
      - Favre filter
      - compute a subset of derived quantities

    Assertions focus on:
      - attributes exist
      - shapes are consistent
      - values are finite
      - a couple of basic physical / numerical sanity checks
    """
    # Setup the folders names
    case_name = "h2_lifted_dns"             # case name
    test_name = "test_pipeline_h2"          # test name

    case_name_orig = case_name + '_orig'    # original dataset name. This one will not be touched by the operations
    dns_data_folder_orig = os.path.join(apriori_test_cache_dir, case_name_orig)

    tests_data_folder = os.path.join(apriori_test_cache_dir, "tmp_tests_results") # The folder where
    dns_data_folder   = os.path.join(tests_data_folder, test_name, case_name)

    os.makedirs(dns_data_folder_orig, exist_ok=True)

    # Download only if needed
    if not os.path.exists(os.path.join(dns_data_folder_orig, "data")):
        ap.download(dataset="h2_premixed", dest_folder=dns_data_folder_orig)

    # Delete the current test folder if it already exists (as it could contain previously processed data), and copy the original one to ensure a clean start
    if os.path.exists(os.path.join(tests_data_folder, test_name)):
        shutil.rmtree(os.path.join(tests_data_folder, test_name))
    else:
        os.makedirs(os.path.join(tests_data_folder, test_name), exist_ok=True)
    shutil.copytree(dns_data_folder_orig, dns_data_folder)

    dns_field = ap.Field(dns_data_folder)

    # --- DNS reaction rates (should create RH2_DNS etc.)
    dns_field.compute_reaction_rates(exist_ok=True)

    # Check a couple of key variables exist and are readable
    assert hasattr(dns_field, "YH2"), "Dataset should expose YH2"
    assert hasattr(dns_field, "RH2_DNS"), "Reaction-rate computation should expose RH2_DNS"

    yh2_dns = dns_field.YH2._3d
    rh2_dns = dns_field.RH2_DNS._3d

    assert list(yh2_dns.shape) == dns_field.shape
    assert list(rh2_dns.shape) == dns_field.shape
    assert np.isfinite(yh2_dns).all()
    assert np.isfinite(rh2_dns).all()

    # Basic sanity: mass fraction should be within [0, 1] up to tiny tolerance
    assert yh2_dns.min() >= -1e-8
    assert yh2_dns.max() <= 1.0 + 1e-8

    # --- Favre filter
    filter_size = 10
    filtered_path = dns_field.filter_favre(filter_size, exist_ok=True)
    filtered_field = ap.Field(filtered_path)

    # Filtered reaction rates
    filtered_field.compute_reaction_rates(exist_ok=True)

    # Check filtered variables exist
    assert hasattr(filtered_field, "YH2")
    assert hasattr(filtered_field, "RH2_DNS")
    assert hasattr(filtered_field, "RH2_LFR"), "LFR rates should be computed by compute_reaction_rates"

    yh2_f = filtered_field.YH2._3d
    rh2_dns_f = filtered_field.RH2_DNS._3d
    rh2_lfr_f = filtered_field.RH2_LFR._3d

    assert list(yh2_f.shape) == filtered_field.shape
    assert np.isfinite(yh2_f).all()
    assert np.isfinite(rh2_dns_f).all()
    assert np.isfinite(rh2_lfr_f).all()

    # Filtering should not explode the range
    assert yh2_f.min() >= -1e-6
    assert yh2_f.max() <= 1.0 + 1e-6

    # --- Derived quantities on filtered field (subset of your script)
    filtered_field.compute_strain_rate(save_tensor=True)
    assert hasattr(filtered_field, "S_LES") or hasattr(filtered_field, "S_DNS"), "strain-rate outputs should exist"

    filtered_field.compute_residual_dissipation_rate(mode="Smag")
    assert hasattr(filtered_field, "Epsilon_r_Smag"), "Smag model output should exist"
    eps = filtered_field.Epsilon_r_Smag._3d
    assert list(eps.shape) == filtered_field.shape
    assert np.isfinite(eps).all()

    filtered_field.compute_residual_kinetic_energy()
    assert hasattr(filtered_field, "K_r_Yosh"), "Residual kinetic energy output attribute 'K_r' should exist "

    # Timescales
    filtered_field.compute_chemical_timescale(mode="SFR", replace_nonreacting="max")
    assert hasattr(filtered_field, "Tau_c_SFR"), "Chemical timescale should exist"
    tau_c = filtered_field.Tau_c_SFR.value
    assert np.isfinite(tau_c).all()

    filtered_field.fuel = "H2"
    filtered_field.ox = "O2"
    filtered_field.compute_chemical_timescale(mode="Ch")
    assert hasattr(filtered_field, "Tau_c_Ch"), "Chomiak timescale should exist"

    filtered_field.compute_mixing_timescale(mode="Kolmo")
    assert hasattr(filtered_field, "Tau_m_Kolmo"), "Mixing timescale should exist"

    # Progress variable + gradient + laplacian(DNS field)
    dns_field.compute_progress_variable("H2O")
    assert hasattr(dns_field, "C"), "Progress variable should exist"

    dns_field.compute_gradient_C()
    assert hasattr(dns_field, "C_grad"), "Progress-variable gradient should exist"

    dns_field.compute_laplacian_C()
    assert hasattr(dns_field, "C_laplacian"), "Progress-variable laplacian should exist"


    # Progress variable + gradient + laplacian (filtered field)
    filtered_field.compute_progress_variable("H2O")
    assert hasattr(filtered_field, "C"), "Progress variable should exist"

    filtered_field.compute_gradient_C()
    assert hasattr(filtered_field, "C_grad"), "Progress-variable gradient should exist"

    filtered_field.compute_laplacian_C()
    assert hasattr(filtered_field, "C_laplacian"), "Progress-variable laplacian should exist"

    # --- Optional: midplane cut smoke test (fast check it doesn't crash)
    cut_field = ap.Field(filtered_field.cut([0, 0, 100], exist_ok=True))
    assert cut_field.shape[2] == 1 or cut_field.shape[2] < filtered_field.shape[2], "Cut should reduce domain"

    # --- Optional: plot figures for qualitative check
    def _save_and_check_fig(fig, outpath: str, dpi: int = 150):
        """
        Save a Matplotlib figure and assert it exists and is non-empty.
        """
        # Ensure folder exists
        os.makedirs(os.path.dirname(outpath), exist_ok=True)

        fig.savefig(outpath, bbox_inches="tight", dpi=dpi)
        # Close to avoid memory build-up during test runs
        try:
            import matplotlib.pyplot as plt
            plt.close(fig)
        except Exception:
            pass

        assert os.path.exists(outpath), f"Expected figure file not found: {outpath}"
        # 10 KB threshold: avoids passing with a zero-byte file
        assert os.path.getsize(outpath) > 10_000, f"Figure seems too small/corrupt: {outpath}"

    figures_dir = os.path.join(tests_data_folder, test_name, "figures")
    dpi = 200

    # --- Plot: YH2 DNS midplane
    fig1, ax1 = dns_field.plot_z_midplane(
        "YH2",
        transpose=True,
        colormap="hot_r",
        vmin=0.0,
        vmax=0.015,
        title=r"$Y_{H_2}$",
        remove_cbar=False,
    )
    _save_and_check_fig(fig1, os.path.join(figures_dir, "YH2_dns.png"), dpi=dpi)

    # --- Plot: YH2 filtered midplane
    fig2, ax2 = filtered_field.plot_z_midplane(
        "YH2",
        transpose=True,
        colormap="hot_r",
        vmin=0.0,
        vmax=0.015,
        title=r"$\widetilde{Y}_{H_2}$",
        cbar_title=r"$Y_{H_2}$ [-]",
        remove_y=False,
    )
    _save_and_check_fig(fig2, os.path.join(figures_dir, "YH2_filtered.png"), dpi=dpi)
    
    # --- Plot: Progress variable
    fig3, ax3 = dns_field.plot_z_midplane(
        "C",
        transpose=True,
        colormap="hot_r",
        # title=r"$c$",
        cbar_title=r"$c$ [-]",
        remove_y=True,
    )
    _save_and_check_fig(fig3, os.path.join(figures_dir, "C.png"), dpi=dpi)

    # --- Plot: Filtered progress variable 
    fig4, ax4 = filtered_field.plot_z_midplane(
        "C",
        transpose=True,
        colormap="hot_r",
        # title="$\widetilde{c}$",
        cbar_title=r"$\widetilde{c}$",
    )
    _save_and_check_fig(fig4, os.path.join(figures_dir, "C_filtered.png"), dpi=dpi)

    # --- Plot: Progress variable gradient
    fig5, ax5 = dns_field.plot_z_midplane(
        "C_grad",
        transpose=True,
        colormap="hot_r",
        # title=r"$c$",
        cbar_title=r"$c$ [-]",
        remove_y=True,
    )
    _save_and_check_fig(fig5, os.path.join(figures_dir, "C_grad.png"), dpi=dpi)

    # --- Plot: Filtered progress variable 
    fig6, ax6 = filtered_field.plot_z_midplane(
        "C_grad",
        transpose=True,
        colormap="hot_r",
        # title="$\widetilde{c}$",
        cbar_title=r"$\widetilde{c}$",
    )
    _save_and_check_fig(fig6, os.path.join(figures_dir, "C_grad_filtered.png"), dpi=dpi)
    
    # --- Plot: Progress variable
    fig7, ax7 = dns_field.plot_z_midplane(
        "C_laplacian",
        transpose=True,
        colormap="hot_r",
        # title=r"$c$",
        cbar_title=r"$c$ [-]",
        remove_y=True,
    )
    _save_and_check_fig(fig7, os.path.join(figures_dir, "C_laplacian.png"), dpi=dpi)

    # --- Plot: Filtered progress variable 
    fig8, ax8 = filtered_field.plot_z_midplane(
        "C_laplacian",
        transpose=True,
        colormap="hot_r",
        # title="$\widetilde{c}$",
        cbar_title=r"$\widetilde{c}$",
    )
    _save_and_check_fig(fig8, os.path.join(figures_dir, "C_laplacian_filtered.png"), dpi=dpi)
    
